#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Components for command line interface for reading graphs

CNFgen has many command line entry points to its functionality, and
some of them expose the same functionality over and over. This module
contains useful common components.

Copyright (C) 2012, 2013, 2014, 2015, 2016, 2019, 2020  Massimo Lauria <lauria@kth.se>
https://github.com/MassimoLauria/cnfgen.git

"""

import sys
import os
import argparse
import textwrap

from itertools import combinations, product

import random
import networkx

from networkx.algorithms.bipartite import complete_bipartite_graph
from networkx.algorithms.bipartite import random_graph as bipartite_random_graph
from networkx.algorithms.bipartite import gnmk_random_graph as bipartite_gnmk_random_graph

from cnfformula import supported_graph_formats
from cnfformula.graphs import readGraph, writeGraph
from cnfformula.graphs import bipartite_random_left_regular, bipartite_random_regular, bipartite_shift
from cnfformula.graphs import bipartite_sets
from cnfformula.graphs import dag_complete_binary_tree, dag_pyramid
from cnfformula.graphs import sample_missing_edges

from .msg import interactive_msg, msg_prefix

from .cmdline import redirect_stdin, CLIError, positive_int


def read_graph_from_input(args, suffix, grtype):
    """Read a graph from input according to command line arguments

    Parameters:
    -----------
    args:
         result of argparse
    gytype:
         one of {dag,bipartite, simple}
    """
    # read graphs from input
    fsource = getattr(args, "input" + suffix)
    fformat = getattr(args, 'graphformat' + suffix)
    try:
        fext = os.path.splitext(fsource.name)[-1][1:]
    except AttributeError:
        fext = ''
    allowed = supported_graph_formats()[grtype]

    no_ext = """
    The formula generation process you asked for needs a {} graph in
    input. Graph format was not specified on the command line and there no
    file name extension to guess that from, thus it is impossible
    to proceed.""".format(grtype)

    bad_ext = """
    The formula generation process you asked for needs a {} graph in
    input. Graph format was not specified on the command line and the
    file name extension do not corresponds to any of the allowed
    format, thus it is impossible to proceed.""".format(grtype)

    ask_opt = """
    Please add the option '-gf <format>' to the graph specification,
    where <format> in one of {}""".format(allowed)

    ask_gr = "Please insert on <stdin> a simple undirected graph in {} format.".format(
        fformat)

    no_ext = textwrap.dedent(no_ext)
    # no_ext = textwrap.fill(no_ext, width=60)
    bad_ext = textwrap.dedent(bad_ext)
    # bad_ext = textwrap.fill(bad_ext, width=60)

    ask_opt = textwrap.dedent(ask_opt)
    # ask_opt = textwrap.fill(ask_opt, width=60)

    with redirect_stdin(fsource):

        # detect graph format
        if fformat == 'autodetect':
            if len(fext) == 0:
                raise CLIError(no_ext + "\n" + ask_opt)

            elif fext not in allowed:
                raise CLIError(bad_ext + "\n" + ask_opt)

        with msg_prefix("INPUT: "):
            interactive_msg(ask_gr, filltext=70)

        G = readGraph(sys.stdin, grtype, fformat)

    G.name = "{} graph from file '{}' (format: {})".format(
        grtype, fsource.name, fformat)
    return G


# Graph readers/generators
class GraphHelper(object):
    """Command Line helper for reading graphs
    """
    @staticmethod
    def setup_command_line(parser, suffix="", required=False):
        """Setup command line options for getting graphs"""
        raise NotImplementedError("Graph Input helper must be subclassed")

    @staticmethod
    def obtain_graph(args, suffix=""):
        """Read/Generate the graph according to the command line options"""
        raise NotImplementedError("Graph Input helper must be subclassed")


class DirectedAcyclicGraphHelper(GraphHelper):
    @staticmethod
    def setup_command_line(parser, suffix="", required=False):
        """Setup command line options for reading a DAG

        Parameters
        ----------
        parser : ArgParse parser object
            it is populated with options for input graphs

        suffix: string, optional
            add a suffix to all input options. Useful if you need to input
            multiple graphs in the same command line (default: empty)

        require : bool, optional
            enforce that at least one input specification is required.
            If it is not the case the standard input is the default input.
            Not a good idea if we read multiple graphs in input.
        """

        gr = parser.add_argument_group(
            title="Input directed acyclic graph (DAG) " + suffix,
            description="""
                                     You can either read the input DAG from file according to one of
                                     the formats, or generate it using one of the constructions included."""
        )

        gr = gr.add_mutually_exclusive_group(required=required)

        gr.add_argument(
            '--input' + suffix,
            '-i' + suffix,
            type=argparse.FileType('r'),
            metavar="<input>",
            default='-',
            help=
            """Read the DAG from <input>. Setting '<input>' to '-' is another way
                        to read from standard input. (default: -) """)

        gr.add_argument('--tree' + suffix,
                        type=positive_int,
                        action='store',
                        metavar="<height>",
                        help="rooted tree digraph")

        gr.add_argument('--pyramid' + suffix,
                        type=positive_int,
                        action='store',
                        metavar="<height>",
                        help="pyramid digraph")

        gr = parser.add_argument_group("I/O options")

        gr.add_argument('--savegraph' + suffix,
                        '-sg' + suffix,
                        type=argparse.FileType('w'),
                        metavar="<graph_file>",
                        default=None,
                        help="""Save the DAG to <graph_file>.
                            Setting '<graph_file>' to '-' is
                            another way to send the DAG to
                            standard output.  (default:  -)
                            """)
        gr.add_argument('--graphformat' + suffix,
                        '-gf' + suffix,
                        choices=supported_graph_formats()['dag'],
                        default='autodetect',
                        help="Format of the DAG file. (default: autodetect)")

    @staticmethod
    def obtain_graph(args, suffix=""):
        """Produce a DAG from either input or library
        """
        if getattr(args, 'tree' + suffix) is not None:
            assert getattr(args, 'tree' + suffix) > 0

            h = getattr(args, 'tree' + suffix)
            G = dag_complete_binary_tree(h)
            G.name = "Complete binary tree of height {}".format(h)

        elif getattr(args, 'pyramid' + suffix) is not None:
            assert getattr(args, 'pyramid' + suffix) > 0

            h = getattr(args, 'pyramid' + suffix)
            G = dag_pyramid(h)
            G.name = "Pyramid graph of height {}".format(h)

        else:
            G = read_graph_from_input(args, suffix, "dag")

        # Output the graph is requested
        if getattr(args, 'savegraph' + suffix) is not None:
            writeGraph(G, getattr(args, 'savegraph' + suffix), "dag",
                       getattr(args, 'graphformat' + suffix))

        return G


class SimpleGraphHelper(GraphHelper):
    @staticmethod
    def setup_command_line(parser, suffix="", required=False):
        """Setup input options for command lines

        Parameters
        ----------
        parser : ArgParse parser object
            it is populated with options for input graphs

        suffix: string, optional
            add a suffix to all input options. Useful if you need to input
            multiple graphs in the same command line (default: empty)

        require : bool, optional
            enforce that at least one input specification is required.
            If it is not the case the standard input is the default input.
            Not a good idea if we read multiple graphs in input.
        """

        gr = parser.add_argument_group(title="Input graph " + suffix,
                                       description="""
                                     You can either read the input graph from file according to one of
                                     the formats, or generate it using one of the included constructions."""
                                       )

        class IntFloat(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                try:
                    n, p = positive_int(values[0]), float(values[1])
                    if p > 1.0 or p < 0:
                        raise ValueError('p must be a float between 0 and 1')
                except ValueError as e:
                    parser.error(str(e))
                setattr(args, self.dest, (n, p))

        gr = gr.add_mutually_exclusive_group(required=required)

        gr.add_argument('--input' + suffix,
                        '-i' + suffix,
                        type=argparse.FileType('r'),
                        metavar="<input>",
                        default='-',
                        help="""Read the graph from <input>.
                        Setting '<input>' to '-' reads the graph from standard
                        input.  (default: -)
                        """)

        gr.add_argument(
            '--gnp' + suffix,
            nargs=2,
            action=IntFloat,
            metavar=('n', 'p'),
            help=
            "random graph according to G(n,p) model (i.e. independent edges)")

        gr.add_argument(
            '--gnm' + suffix,
            type=positive_int,
            nargs=2,
            action='store',
            metavar=('n', 'm'),
            help="random graph according to G(n,m) model (i.e. m random edges)"
        )

        gr.add_argument(
            '--gnd' + suffix,
            type=positive_int,
            nargs=2,
            action='store',
            metavar=('n', 'd'),
            help=
            "random d-regular graph according to G(n,d) model (i.e. d random edges per vertex)"
        )

        gr.add_argument('--grid' + suffix,
                        type=positive_int,
                        nargs='+',
                        action='store',
                        metavar=('d1', 'd2'),
                        help="n-dimensional grid of dimension d1 x d2 x ... ")

        gr.add_argument(
            '--torus' + suffix,
            type=positive_int,
            nargs='+',
            action='store',
            metavar=('d1', 'd2'),
            help="n-dimensional torus grid of dimensions d1 x d2 x ... x dn")

        gr.add_argument('--complete' + suffix,
                        type=positive_int,
                        action='store',
                        metavar="<N>",
                        help="complete graph on N vertices")

        gr.add_argument('--empty' + suffix,
                        type=positive_int,
                        action='store',
                        metavar="<N>",
                        help="empty graph on N vertices")

        gr = parser.add_argument_group("Modifications for input graph " +
                                       suffix)
        gr.add_argument(
            '--plantclique' + suffix,
            type=positive_int,
            action='store',
            metavar="<k>",
            help="choose k vertices at random and add all edges among them")

        gr.add_argument(
            '--addedges' + suffix,
            type=positive_int,
            action='store',
            metavar="<k>",
            help="add k NEW random edges to the graph (applied last)")

        gr = parser.add_argument_group("I/O options for graph " + suffix)
        gr.add_argument('--savegraph' + suffix,
                        '-sg' + suffix,
                        type=argparse.FileType('w'),
                        metavar="<graph_file>",
                        default=None,
                        help="""Save the graph to <graph_file>.
                            Setting '<graph_file>' to '-' is
                            another way to send the graph to
                            standard output.  (default:  -)
                            """)
        gr.add_argument('--graphformat' + suffix,
                        '-gf' + suffix,
                        choices=supported_graph_formats()['simple'],
                        default='autodetect',
                        help="Format of the graph file. (default: autodetect)")

    @staticmethod
    def obtain_graph(args, suffix=""):
        """Build a Graph according to command line arguments

        Arguments:
        - `args`: command line options
        """
        if getattr(args, 'gnd' + suffix) is not None:

            n, d = getattr(args, 'gnd' + suffix)
            if (n * d) % 2 == 1:
                raise CLIError("ERROR: n * d must be even")
            G = networkx.random_regular_graph(d, n)
            G.name = 'Random {}-regular graph of {} vertices'.format(d, n)

        elif getattr(args, 'gnp' + suffix) is not None:

            n, p = getattr(args, 'gnp' + suffix)
            G = networkx.gnp_random_graph(n, p)
            G.name = 'Random {}-biased graph of {} vertices'.format(p, n)

        elif getattr(args, 'gnm' + suffix) is not None:

            n, m = getattr(args, 'gnm' + suffix)
            G = networkx.gnm_random_graph(n, m)
            G.name = 'Random graph of {} vertices with {} edges'.format(n, m)

        elif getattr(args, 'grid' + suffix) is not None:

            dim = getattr(args, 'grid' + suffix)
            G = networkx.grid_graph(dim)
            G.name = "Grid graph of dimension {}".format(dim)

        elif getattr(args, 'torus' + suffix) is not None:

            dim = getattr(args, 'torus' + suffix)
            G = networkx.grid_graph(dim, periodic=True)
            G.name = "Torus graph of dimension {}".format(dim)

        elif getattr(args, 'complete' + suffix) is not None:

            n = getattr(args, 'complete' + suffix)
            G = networkx.complete_graph(n)
            G.name = "Complete graphs of {} vertices".format(n)

        elif getattr(args, 'empty' + suffix) is not None:

            n = getattr(args, 'empty' + suffix)
            G = networkx.empty_graph(n)
            G.name = 'Empty graph of {} vertices'.format(n)

        else:

            G = read_graph_from_input(args, suffix, "simple")

        # Graph modifications
        if getattr(args, 'plantclique' + suffix) is not None and getattr(
                args, 'plantclique' + suffix) > 1:
            cliquesize = getattr(args, 'plantclique' + suffix)
            if cliquesize > G.order():
                raise ValueError("Clique cannot be larger than graph")

            clique = random.sample(G.nodes(), cliquesize)

            for v, w in combinations(clique, 2):
                G.add_edge(v, w)
            G.name += " + planted {}-clique".format(cliquesize)

        if getattr(args, 'addedges' + suffix) is not None and getattr(
                args, 'addedges' + suffix) > 0:
            k = getattr(args, 'addedges' + suffix)
            G.add_edges_from(sample_missing_edges(G, k))
            G.name += " + {} random edges".format(k)

        # Output the graph is requested
        if getattr(args, 'savegraph' + suffix) is not None:
            writeGraph(G, getattr(args, 'savegraph' + suffix), 'simple',
                       getattr(args, 'graphformat' + suffix))

        assert hasattr(G, 'name')
        return G


class BipartiteGraphHelper(GraphHelper):
    @staticmethod
    def setup_command_line(parser, suffix="", required=False):
        """Setup input options for reading bipartites on command lines

        Parameters
        ----------
        parser : ArgParse parser object
            it is populated with options for input graphs

        suffix: string, optional
            add a suffix to all input options. Useful if you need to input
            multiple graphs in the same command line (default: empty)

        require : bool, optional
            enforce that at least one input specification is required.
            If it is not the case the standard input is the default input.
            Not a good idea if we read multiple graphs in input.
        """
        class IntIntFloat(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                try:
                    l, r, p = positive_int(values[0]), positive_int(
                        values[1]), float(values[2])
                    if not 0.0 <= p <= 1.0:
                        raise ValueError('p must be a float between 0 and 1')
                except ValueError as e:
                    raise argparse.ArgumentError(self, str(e))
                setattr(args, self.dest, (l, r, p))

        class BipartiteRegular(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                try:
                    l, r, d = positive_int(values[0]), positive_int(
                        values[1]), positive_int(values[2])
                    if d > r:
                        raise ValueError(
                            'In a regular bipartite graph, left degree d is at most r.'
                        )
                    if (d * l % r) != 0:
                        raise ValueError(
                            'In a regular bipartite graph, r must divide d*l.')
                except ValueError as e:
                    raise argparse.ArgumentError(self, str(e))
                setattr(args, self.dest, (l, r, d))

        class BipartiteEdge(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                try:
                    l, r, m = positive_int(values[0]), positive_int(
                        values[1]), positive_int(values[2])
                    if m > r * l:
                        raise ValueError(
                            'In a bipartite graph, #edges is at most l*r.')
                except ValueError as e:
                    raise argparse.ArgumentError(self, str(e))
                setattr(args, self.dest, (l, r, m))

        class BipartiteShift(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                try:

                    if len(values) < 2:
                        raise ValueError(
                            "'bshift' requires two positive int parameters or more."
                        )

                    N, M, pattern = values[0], values[1], sorted(values[2:])

                    for i in range(len(pattern) - 1):
                        if pattern[i] == pattern[i + 1]:
                            raise ValueError(
                                "no repetitions is allowed in the edge pattern."
                            )

                    if N < 1 or M < 1:
                        raise ValueError(
                            "matrix dimensions N and M must be positive.")

                    if any([x < 1 or x > M for x in pattern]):
                        raise ValueError(
                            "in v(1),v(2)... we need 1 <= v(i) <= M.")

                except ValueError as e:
                    raise argparse.ArgumentError(self, str(e))
                setattr(args, self.dest, (N, M, pattern))

        class BipartiteLeft(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                try:
                    l, r, d = positive_int(values[0]), positive_int(
                        values[1]), positive_int(values[2])
                    if d > r:
                        raise ValueError(
                            'In a bipartite graph, left degree d is at most r.'
                        )
                except ValueError as e:
                    raise argparse.ArgumentError(self, str(e))
                setattr(args, self.dest, (l, r, d))

        gr = parser.add_argument_group("Bipartite graph structure " + suffix,
                                       description="""
                                     The structure of this CNF formula depends on a bipartite graph, which
                                     can be read from file (in one of the supported format), or generated
                                     using one of the included constructions."""
                                       )

        gr = gr.add_mutually_exclusive_group(required=False)

        gr.add_argument(
            '--input' + suffix,
            '-i' + suffix,
            type=argparse.FileType('r'),
            metavar="<input>",
            default='-',
            help="""Read the graph from file. Setting '<input>' to '-' is
                        another way to read from standard input.  (default: -)
                        """)

        gr.add_argument('--bp' + suffix,
                        nargs=3,
                        action=IntIntFloat,
                        metavar=('l', 'r', 'p'),
                        help="Random bipartite graph with independent edges")

        gr.add_argument('--bm' + suffix,
                        type=positive_int,
                        nargs=3,
                        action=BipartiteEdge,
                        metavar=('l', 'r', 'm'),
                        help="Bipartite graph with m random edges")

        gr.add_argument(
            '--bd' + suffix,
            type=positive_int,
            nargs=3,
            action=BipartiteLeft,
            metavar=('l', 'r', 'd'),
            help="Bipartite graph with d random edges per left vertex")

        gr.add_argument(
            '--bregular' + suffix,
            nargs=3,
            action=BipartiteRegular,
            metavar=('l', 'r', 'd'),
            help="Bipartite regular graph, with d random edges per left vertex."
        )

        gr.add_argument(
            '--bshift' + suffix,
            type=positive_int,
            nargs='*',
            action=BipartiteShift,
            metavar=('N'),
            help=
            "Args <N> <M> <v1> <v2> ... NxM bipartite. Vertex i connexted to i+v1, i+v2,... (mod M)"
        )

        gr.add_argument('--bcomplete' + suffix,
                        type=positive_int,
                        nargs=2,
                        action='store',
                        metavar=('l', 'r'),
                        help="Complete bipartite graph")

        gr = parser.add_argument_group("Modify the graph structure")

        gr.add_argument('--plantbiclique' + suffix,
                        type=positive_int,
                        nargs=2,
                        action='store',
                        metavar=('l', 'r'),
                        help="Plant a random (l,r)-bipartite clique")

        gr.add_argument(
            '--addedges' + suffix,
            type=positive_int,
            action='store',
            metavar="<k>",
            help="Add k NEW random edges to the graph (applied in the end)")

        gr = parser.add_argument_group("File I/O options",
                                       description="""
                                     Additional option regarding the input and output of the files
                                     containing the graph structure.
                                     """)
        gr.add_argument(
            '--savegraph' + suffix,
            '-sg' + suffix,
            type=argparse.FileType('w'),
            metavar="<graph_file>",
            default=None,
            help=
            """Save the graph to <graph_file>. Setting '<graph_file>' to '-'sends
                        the graph to standard output. (default: -) """)
        gr.add_argument('--graphformat' + suffix,
                        '-gf' + suffix,
                        choices=supported_graph_formats()['bipartite'],
                        default='autodetect',
                        help="Format of the graph file. (default: autodetect)")

    @staticmethod
    def obtain_graph(args, suffix=""):
        """Build a Bipartite graph according to command line arguments

        Arguments:
        - `args`: command line options
        """

        if getattr(args, "bp" + suffix) is not None:

            l, r, p = getattr(args, "bp" + suffix)
            G = bipartite_random_graph(l, r, p)
            G.name = "Random {}-biased bipartite with {},{} vertices".format(
                p, l, r)

        elif getattr(args, "bm" + suffix) is not None:

            l, r, m = getattr(args, "bm" + suffix)
            G = bipartite_gnmk_random_graph(l, r, m)
            G.name = "Random bipartite with {},{} vertices and {} edges".format(
                l, r, m)

        elif getattr(args, "bd" + suffix) is not None:

            l, r, d = getattr(args, "bd" + suffix)
            G = bipartite_random_left_regular(l, r, d)
            G.name = "Random {}-left regular bipartite with {},{} vertices".format(
                d, l, r)

        elif getattr(args, "bregular" + suffix) is not None:

            l, r, d = getattr(args, "bregular" + suffix)
            G = bipartite_random_regular(l, r, d)
            G.name = "Random {}-regular bipartite with {},{} vertices".format(
                d, l, r)

        elif getattr(args, "bshift" + suffix) is not None:

            N, M, pattern = getattr(args, "bshift" + suffix)
            G = bipartite_shift(N, M, pattern)
            G.name = 'Bipartite with {},{} vertices and shifting edge pattern {}'.format(
                N, M, pattern)

        elif getattr(args, "bcomplete" + suffix) is not None:

            l, r = getattr(args, "bcomplete" + suffix)
            G = complete_bipartite_graph(l, r)
            G.name = "Complete bipartite with {},{} vertices".format(l, r)

        else:
            G = read_graph_from_input(args, suffix, "bipartite")

        # Graph modifications
        if getattr(args, "plantbiclique" + suffix) is not None:
            l, r = getattr(args, "plantbiclique" + suffix)
            left, right = bipartite_sets(G)
            if l > len(left) or r > len(right):
                raise ValueError("Clique cannot be larger than graph")

            left = random.sample(left, l)
            right = random.sample(right, r)

            for v, w in product(left, right):
                G.add_edge(v, w)

            G.name += " + planted {}x{}-biclique".format(l, r)

        if getattr(args, "addedges" + suffix) is not None:

            k = getattr(args, "addedges" + suffix)

            G.add_edges_from(sample_missing_edges(G, k))

            G.name += " + {} random edges".format(k)

        # Output the graph is requested
        if getattr(args, "savegraph" + suffix) is not None:
            writeGraph(G, getattr(args, "savegraph" + suffix), 'bipartite',
                       getattr(args, "graphformat" + suffix))

        assert hasattr(G, 'name')
        return G



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Formula Helpers for reading dimacs

Copyright (C) 2012-2020 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""

import argparse
from cnfformula import readCNF

from .formula_helpers import FormulaHelper
from .cmdline import redirect_stdin
from .msg import interactive_msg
from .msg import msg_prefix

usage_string = """Dimacs reader

This is not a real formula construction. The formula is built by
reading a DIMACS file. This command is particularly useful to apply
some transformations to that file.

examples:
 {0}           --- read dimacs CNF from standard input
 {0} file.cnf  --- read dimacs CNF from file.cnf
"""


class DimacsCmdHelper(FormulaHelper):
    """Command line helper for the Pigeonhole principle CNF"""

    name = 'dimacs'
    description = 'Read dimacs file'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for pigeonhole principle formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.usage = usage_string.format(parser.prog)
        parser.add_argument('input',
                            nargs='?',
                            help=argparse.SUPPRESS,
                            type=argparse.FileType('r'),
                            default='-')

    @staticmethod
    def build_cnf(args):

        msg = """Waiting for a DIMACS formula on <stdin>.
Alternatively you can feed a formula to <stdin>
with piping or giving a filename on the command line."""

        with redirect_stdin(args.input):
            with msg_prefix("INPUT: "):
                interactive_msg(msg, filltext=70)

            F = readCNF()

        return F



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of the ordering principle formulas
"""

from cnfformula.cnf import CNF

from cnfformula.graphs import enumerate_vertices, neighbors
from itertools import combinations, permutations

import networkx


def OrderingPrinciple(size, total=False, smart=False, plant=False, knuth=0):
    """Generates the clauses for ordering principle

    Arguments:
    - `size`  : size of the domain
    - `total` : add totality axioms (i.e. "x < y" or "x > y")
    - `smart` : "x < y" and "x > y" are represented by a single variable (implies totality)
    - `plant` : allow a single element to be minimum (could make the formula SAT)
    - `knuth` : Donald Knuth variant of the formula ver. 2 or 3 (anything else suppress it)
    """
    if total or smart:
        description = "Total ordering principle"
    else:
        description = "Ordering principle"

    if smart:
        description += " (compact representation)"

    if knuth in [2, 3]:
        description += " (Knuth variant {})".format(knuth)

    F = GraphOrderingPrinciple(networkx.complete_graph(size), total, smart,
                               plant, knuth)
    F.header['description'] = description
    return F


def varname(v1, v2):
    return 'x_{{{0},{1}}}'.format(v1, v2)


def GraphOrderingPrinciple(graph,
                           total=False,
                           smart=False,
                           plant=False,
                           knuth=0):
    """Generates the clauses for graph ordering principle

    Arguments:
    - `graph` : undirected graph
    - `total` : add totality axioms (i.e. "x < y" or "x > y")
    - `smart` : "x < y" and "x > y" are represented by a single variable (implies `total`)
    - `plant` : allow last element to be minimum (and could make the formula SAT)
    - `knuth` : Don Knuth variants 2 or 3 of the formula (anything else suppress it)
    """
    # Describe the formula
    if total or smart:
        description = "Total graph ordering principle"
    else:
        description = "Graph ordering principle"

    if smart:
        description += " (compact representation)"

    if knuth in [2, 3]:
        description += " (Knuth variant {})".format(knuth)

    if hasattr(graph, 'name'):
        description += " on " + graph.name

    gop = CNF(description=description)

    # Fix the vertex order
    V = enumerate_vertices(graph)

    # Add variables
    iterator = combinations if smart else permutations
    for v1, v2 in iterator(V, 2):
        gop.add_variable(varname(v1, v2))

    #
    # Non minimality axioms
    #

    # Clause is generated in such a way that if totality is enforces,
    # every pair occurs with a specific orientation.
    # Allow minimum on last vertex if 'plant' options.

    for med in range(len(V) - (plant and 1)):
        clause = []
        for lo in range(med):
            if graph.has_edge(V[med], V[lo]):
                clause += [(True, varname(V[lo], V[med]))]
        for hi in range(med + 1, len(V)):
            if not graph.has_edge(V[med], V[hi]):
                continue
            elif smart:
                clause += [(False, varname(V[med], V[hi]))]
            else:
                clause += [(True, varname(V[hi], V[med]))]
        gop.add_clause(clause, strict=True)

    #
    # Transitivity axiom
    #

    if len(V) >= 3:
        if smart:
            # Optimized version if smart representation of totality is used
            for (v1, v2, v3) in combinations(V, 3):

                gop.add_clause([(True, varname(v1, v2)),
                                (True, varname(v2, v3)),
                                (False, varname(v1, v3))],
                               strict=True)

                gop.add_clause([(False, varname(v1, v2)),
                                (False, varname(v2, v3)),
                                (True, varname(v1, v3))],
                               strict=True)

        elif total:
            # With totality we still need just two axiom per triangle
            for (v1, v2, v3) in combinations(V, 3):

                gop.add_clause([(False, varname(v1, v2)),
                                (False, varname(v2, v3)),
                                (False, varname(v3, v1))],
                               strict=True)

                gop.add_clause([(False, varname(v1, v3)),
                                (False, varname(v3, v2)),
                                (False, varname(v2, v1))],
                               strict=True)

        else:
            for (v1, v2, v3) in permutations(V, 3):

                # knuth variants will reduce the number of
                # transitivity axioms
                if knuth == 2 and ((v2 < v1) or (v2 < v3)):
                    continue
                if knuth == 3 and ((v3 < v1) or (v3 < v2)):
                    continue

                gop.add_clause([(False, varname(v1, v2)),
                                (False, varname(v2, v3)),
                                (True, varname(v1, v3))],
                               strict=True)

    if not smart:
        # Antisymmetry axioms (useless for 'smart' representation)
        for (v1, v2) in combinations(V, 2):
            gop.add_clause([(False, varname(v1, v2)),
                            (False, varname(v2, v1))],
                           strict=True)

        # Totality axioms (useless for 'smart' representation)
        if total:
            for (v1, v2) in combinations(V, 2):
                gop.add_clause([(True, varname(v1, v2)),
                                (True, varname(v2, v1))],
                               strict=True)

    return gop



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Transformation Helpers for command line

Copyright (C) 2012, 2013, 2014, 2015, 2016, 2019, 2020 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""

# Formula transformation implemented
from cnfformula import Shuffle
from cnfformula import AllEqualSubstitution
from cnfformula import ExactlyOneSubstitution
from cnfformula import FlipPolarity
from cnfformula import FormulaLifting
from cnfformula import IfThenElseSubstitution
from cnfformula import MajoritySubstitution
from cnfformula import NotAllEqualSubstitution
from cnfformula import OrSubstitution
from cnfformula import VariableCompression
from cnfformula import XorSubstitution

from .graph_cmdline import BipartiteGraphHelper


class TransformationHelper:
    """Command line helper for a formula family"""
    name = None
    description = None

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line parser for this transformation subcommand"""
        raise NotImplementedError(
            "Transformation family helper must be subclassed")

    @staticmethod
    def transform_cnf(F, args):
        """Build the new CNF by applying the transformation"""
        raise NotImplementedError(
            "Transformation family helper must be subclassed")


class ShuffleCmd(TransformationHelper):
    """Shuffle 
    """
    name = 'shuffle'
    description = 'Permute variables, clauses and polarity of literals at random'

    @staticmethod
    def setup_command_line(parser):
        parser.add_argument('--no-polarity-flips',
                            '-p',
                            action='store_true',
                            default=False,
                            dest='no_polarity_flips',
                            help="No polarity flips")
        parser.add_argument('--no-variables-permutation',
                            '-v',
                            action='store_true',
                            default=False,
                            dest='no_variables_permutation',
                            help="No permutation of variables")
        parser.add_argument('--no-clauses-permutation',
                            '-c',
                            action='store_true',
                            default=False,
                            dest='no_clauses_permutation',
                            help="No permutation of clauses")

    @staticmethod
    def transform_cnf(F, args):
        return Shuffle(
            F,
            variables_permutation=None
            if not args.no_variables_permutation else list(F.variables()),
            clauses_permutation=None
            if not args.no_clauses_permutation else list(range(len(F))),
            polarity_flips=None if not args.no_polarity_flips else [1] *
            len(list(F.variables())))


#
# Command line helpers for these substitutions
#


class NoSubstitutionCmd(TransformationHelper):
    name = 'none'
    description = 'no transformation'

    @staticmethod
    def setup_command_line(parser):
        pass

    @staticmethod
    def transform_cnf(F, args):
        return F


class OrSubstitutionCmd(TransformationHelper):
    name = 'or'
    description = 'substitute variable x with OR(x1,x2,...,xN)'

    @staticmethod
    def setup_command_line(parser):
        parser.add_argument('N',
                            type=int,
                            nargs='?',
                            default=2,
                            action='store',
                            help="arity (default: 2)")

    @staticmethod
    def transform_cnf(F, args):
        return OrSubstitution(F, args.N)


class XorSubstitutionCmd(TransformationHelper):
    name = 'xor'
    description = 'substitute variable x with XOR(x1,x2,...,xN)'

    @staticmethod
    def setup_command_line(parser):
        parser.add_argument('N',
                            type=int,
                            nargs='?',
                            default=2,
                            action='store',
                            help="arity (default: 2)")

    @staticmethod
    def transform_cnf(F, args):
        return XorSubstitution(F, args.N)


class AllEqualsSubstitutionCmd(TransformationHelper):
    name = 'eq'
    description = 'substitute variable x with predicate x1==x2==...==xN (i.e. all equals)'

    @staticmethod
    def setup_command_line(parser):
        parser.add_argument('N',
                            type=int,
                            nargs='?',
                            default=3,
                            action='store',
                            help="arity (default: 3)")

    @staticmethod
    def transform_cnf(F, args):
        return AllEqualSubstitution(F, args.N)


class NeqSubstitutionCmd(TransformationHelper):
    name = 'neq'
    description = 'substitute variable x with predicate |{x1,x2,...,xN}|>1 (i.e. not all equals)'

    @staticmethod
    def setup_command_line(parser):
        parser.add_argument('N',
                            type=int,
                            nargs='?',
                            default=3,
                            action='store',
                            help="arity (default: 3)")

    @staticmethod
    def transform_cnf(F, args):
        return NotAllEqualSubstitution(F, args.N)


class MajSubstitution(TransformationHelper):
    name = 'maj'
    description = 'substitute variable x with predicate Majority(x1,x2,...,xN)'

    @staticmethod
    def setup_command_line(parser):
        parser.add_argument('N',
                            type=int,
                            nargs='?',
                            default=3,
                            action='store',
                            help="arity (default: 3)")

    @staticmethod
    def transform_cnf(F, args):
        return MajoritySubstitution(F, args.N)


class IfThenElseSubstitutionCmd(TransformationHelper):
    name = 'ite'
    description = 'substitute variable x with predicate "if X then Y else Z"'

    @staticmethod
    def setup_command_line(parser):
        pass

    @staticmethod
    def transform_cnf(F, args):
        return IfThenElseSubstitution(F)


class ExactlyOneSubstitutionCmd(TransformationHelper):
    name = 'one'
    description = 'substitute variable x with predicate x1+x2+...+xN = 1'

    @staticmethod
    def setup_command_line(parser):
        parser.add_argument('N',
                            type=int,
                            nargs='?',
                            default=3,
                            action='store',
                            help="arity (default: 3)")

    @staticmethod
    def transform_cnf(F, args):
        return ExactlyOneSubstitution(F, args.N)


# Technically lifting is not a substitution, therefore it should be in
# another file. Unfortunately there is a lot of dependency from
# this one.


class FormulaLiftingCmd(TransformationHelper):
    """Lifting 
    """
    name = 'lift'
    description = 'one dimensional lifting  x -->  x1 y1  OR ... OR xN yN, with y1+..+yN = 1'

    @staticmethod
    def setup_command_line(parser):
        parser.add_argument('N',
                            type=int,
                            nargs='?',
                            default=3,
                            action='store',
                            help="arity (default: 3)")

    @staticmethod
    def transform_cnf(F, args):
        return FormulaLifting(F, args.N)


class FlipCmd(TransformationHelper):
    name = 'flip'
    description = 'negate all variables in the formula'

    @staticmethod
    def setup_command_line(parser):
        pass

    @staticmethod
    def transform_cnf(F, args):

        return FlipPolarity(F)


class XorCompressionCmd(TransformationHelper):
    name = 'xorcomp'
    description = 'variable compression using XOR'

    @staticmethod
    def setup_command_line(parser):
        BipartiteGraphHelper.setup_command_line(parser)

    @staticmethod
    def transform_cnf(F, args):
        B = BipartiteGraphHelper.obtain_graph(args)

        return VariableCompression(F, B, function='xor')


class MajCompressionCmd(TransformationHelper):
    name = 'majcomp'
    description = 'variable compression using Majority'

    @staticmethod
    def setup_command_line(parser):
        BipartiteGraphHelper.setup_command_line(parser)

    @staticmethod
    def transform_cnf(F, args):
        B = BipartiteGraphHelper.obtain_graph(args)

        return VariableCompression(F, B, function='maj')



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Formula Helpers for simple and random formulas

Copyright (C) 2012, 2013, 2014, 2015, 2016, 2019, 2020 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""

from cnfformula import CNF
from cnfformula import RandomKCNF

from .formula_helpers import FormulaHelper


class OR(FormulaHelper):
    """Command line helper for a single clause formula
    """

    name = 'or'
    description = 'a single disjunction'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for single or of literals

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('P',
                            metavar='<P>',
                            type=int,
                            help="positive literals")
        parser.add_argument('N',
                            metavar='<N>',
                            type=int,
                            help="negative literals")

    @staticmethod
    def build_cnf(args):
        """Build an disjunction

        Arguments:
        - `args`: command line options
        """
        clause = [ (True,"x_{}".format(i)) for i in range(args.P) ] + \
                 [ (False,"y_{}".format(i)) for i in range(args.N) ]
        description = "Clause with {} positive and {} negative literals".format(
            args.P, args.N)
        orcnf = CNF([clause], description)

        return orcnf


class AND(FormulaHelper):
    """Command line helper for a 1-CNF (i.e. conjunction)
    """
    name = 'and'
    description = 'a single conjunction'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for an and of literals

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('P',
                            metavar='<P>',
                            type=int,
                            help="positive literals")
        parser.add_argument('N',
                            metavar='<N>',
                            type=int,
                            help="negative literals")

    @staticmethod
    def build_cnf(args):
        """Build a conjunction

        Arguments:
        - `args`: command line options
        """
        clauses = [ [(True,"x_{}".format(i))] for i in range(args.P) ] + \
                  [ [(False,"y_{}".format(i))] for i in range(args.N) ]
        andcnf = CNF(clauses)
        andcnf.header = "Singleton clauses: {} positive and {} negative\n\n""".format(args.P,args.N) +\
                        andcnf.header
        return andcnf


class EMPTY(FormulaHelper):
    """Command line helper for the empty CNF (no clauses)
    """

    name = 'empty'
    description = 'empty CNF formula'

    @staticmethod
    def setup_command_line(parser):
        pass

    @staticmethod
    def build_cnf(args):
        """Build an empty CNF formula 

        Parameters
        ----------
        args : ignored 
             command line options
        """
        return CNF()


class EMPTY_CLAUSE(FormulaHelper):
    """Command line helper for the contradiction (one empty clause)  
    """

    name = 'emptyclause'
    description = 'one empty clause'

    @staticmethod
    def setup_command_line(parser):
        pass

    @staticmethod
    def build_cnf(args):
        """Build a CNF formula with an empty clause 

        Parameters
        ----------
        args : ignored 
             command line options
        """
        return CNF([[]])


class RandCmdHelper(FormulaHelper):
    """Command line helper for random formulas
    """
    name = 'randkcnf'
    description = 'random k-CNF'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for an and of literals

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('k', metavar='<k>', type=int, help="clause width")
        parser.add_argument('n',
                            metavar='<n>',
                            type=int,
                            help="number of variables")
        parser.add_argument('m',
                            metavar='<m>',
                            type=int,
                            help="number of clauses")

    @staticmethod
    def build_cnf(args):
        """Build a conjunction

        Arguments:
        - `args`: command line options
        """
        return RandomKCNF(args.k, args.n, args.m)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from cnfformula.cnf import CNF

import cnfformula.families

from itertools import product

import networkx
from cnfformula.families.tseitin import TseitinFormula
from itertools import combinations


def PitfallFormula(v, d, ny, nz, k):
    """Pitfall Formula

    The Pitfall formula was designed to be specifically easy for
    Resolution and hard for common CDCL heuristics. The formula is
    unsatisfiable and consists of three parts: an easy formula, a hard
    formula, and a pitfall misleading the solver into working with the
    hard part.

    The hard part are several copies of an unsatisfiable Tseitin
    formula on a random regular graph. The pitfall part is made up of
    a few gadgets over (primarily) two sets of variables: pitfall
    variables, which point the solver towards the hard part after
    being assigned, and safety variables, which prevent the gadget
    from breaking even if a few other variables are assigned.

    For more details, see the corresponding paper [1]_.

    Parameters
    ----------
    v : positive integer
        number of vertices of the Tseitin graph
    d : positive integer
        degree of the Tseitin graph  
    ny : positive integer
         number of pitfall variables
    nz : positive integer
         number of safety variables
    k : positive, even integer
        number of copies of the hard and pitfall parts; controls how
        easy the easy part is

    Returns
    -------
    A CNF object

    Raises
    ------
    ValueError
        When `v < d` or `d*v` is odd there is no d-regular graph on `v` verices.

    References
    ----------
    .. [1] Marc Vinyals.
           Hard examples for common variable decision heuristics.
           In, AAAI 2020 (pp. 1652–1659).
    """
    if v <= 0:
        raise ValueError("v must be positive.")
    if d <= 0:
        raise ValueError("d must be positive.")
    if ny <= 0:
        raise ValueError("ny must be positive.")
    if nz <= 0:
        raise ValueError("nz must be positive.")
    if k <= 0:
        raise ValueError("k must be positive.")
    if k % 2 != 0:
        raise ValueError("k must be even.")

    def xname(j, x):
        return "{}_{}".format(x, j)

    phi = CNF(
        description=
        'Pitfall Formula with parameters (v={},d={},ny={},nz={},k={})'.format(
            v, d, ny, nz, k))
    try:
        graph = networkx.random_regular_graph(d, v)
    except networkx.exception.NetworkXError:
        raise ValueError("""No regular {}-degree graph with {}-vertices exists.
Degree d must less than the number v of vertices,
and d*v must be even.""".format(d, v))

    charge = [1] + [0] * (v - 1)
    ts = TseitinFormula(graph, charge)

    X_ = list(ts.variables())
    nx = len(X_)

    ### Variables
    X = [0] * k
    P = [0] * k
    Y = [0] * k
    Z = [0] * k
    A = [0] * k
    for j in range(k):
        X[j] = [xname(j, x) for x in X_]
        P[j] = ["p_{}_{}".format(j, i) for i in range(nx + nz)]
        Y[j] = ["y_{}_{}".format(j, i) for i in range(ny)]
        Z[j] = ["z_{}_{}".format(j, i) for i in range(nz)]
        A[j] = ["a_{}_{}".format(j, i) for i in range(3)]

    for YY in Y:
        for y in YY:
            phi.add_variable(y)

    for XX in X:
        for x in XX:
            phi.add_variable(x)

    ### Hard part
    # Ts_j
    for j in range(k):
        append = [(True, z) for z in Z[j]]
        for C in ts:
            CC = [(p, xname(j, x)) for (p, x) in C]
            phi.add_clause(CC + append)

    ### Pitfall part
    # Psi
    def pitfall(y1, y2, PP):
        CY = [(True, y1), (True, y2)]
        for p in PP:
            phi.add_clause(CY + [(False, p)])

    for j in range(k):
        for (y1, y2) in combinations(Y[j], 2):
            pitfall(y1, y2, P[j])

    # Pi
    def pipe(y, PP, XX, ZZ):
        S = XX + ZZ
        CY = [(True, y)]
        C = []
        for (s, PPP) in zip(S, combinations(PP, len(PP) - 1)):
            CP = [(True, p) for p in PPP]
            CS = C
            if len(CS) + 1 == len(S):
                # C_{m+n} does not contain z_1
                del CS[nx]
            phi.add_clause(CY + CP + CS + [(False, s)])
            C.append((True, s))

    for j in range(k):
        for y in Y[j]:
            pipe(y, P[j], X[j], Z[j])

    # Delta
    def tail(y, z, AA):
        phi.add_clause([(False, AA[0]), (True, AA[2]), (False, z)])
        phi.add_clause([(False, AA[1]), (False, AA[2]), (False, z)])
        phi.add_clause([(True, AA[0]), (False, z), (False, y)])
        phi.add_clause([(True, AA[1]), (False, z), (False, y)])

    for j in range(k):
        for (y, z) in product(Y[j], Z[j]):
            tail(y, z, A[j])

    ### Easy part
    # Gamma
    split_gamma = 2
    for i in range(0, ny, split_gamma):
        phi.add_clause([(False, Y[j][i + ii]) for j in range(k)
                        for ii in range(split_gamma)])

    return phi



#!/usr/bin/env python

import random
from copy import copy

from cnfformula.cnf import CNF


def Shuffle(cnf,
            variables_permutation=None,
            clauses_permutation=None,
            polarity_flips=None):
    """ Reshuffle the given cnf. Returns a formula logically
    equivalent to the input with the following transformations
    applied in order:

    1. Polarity flips. polarity_flip is a {-1,1}^n vector. If the i-th
    entry is -1, all the literals with the i-th variable change its
    sign.

    2. Variable permutations. variable_permutation is a permutation of
    [vars(cnf)]. All the literals with the old i-th variable are
    replaced with the new i-th variable.

    3. Clause permutations. clause_permutation is a permutation of
    [0..m-1]. The resulting clauses are reordered according to the
    permutation.
"""

    # empty cnf
    out = CNF()
    out.header = copy(cnf.header)

    if 'description' in out.header:
        out.header['description'] += " (reshuffled)"

    i = 1
    while 'transformation {}'.format(i) in out.header:
        i += 1
    out.header['transformation {}'.format(i)] = "Formula reshuffling"

    variables = list(cnf.variables())
    N = len(variables)
    M = len(cnf)

    # variable permutation
    if variables_permutation is None:
        variables_permutation = variables
        random.shuffle(variables_permutation)
    else:
        assert len(variables_permutation) == N

    # polarity flip
    if polarity_flips is None:
        polarity_flips = [random.choice([-1, 1]) for x in range(N)]
    else:
        assert len(polarity_flips) == N

    #
    # substitution of variables
    #
    for v in variables_permutation:
        out.add_variable(v)

    substitution = [None] * (2 * N + 1)
    reverse_idx = dict([(v, i) for (i, v) in enumerate(out.variables(), 1)])
    polarity_flips = [None] + polarity_flips

    for i, v in enumerate(cnf.variables(), 1):
        substitution[i] = polarity_flips[i] * reverse_idx[v]
        substitution[-i] = -substitution[i]

    #
    # permutation of clauses
    #
    if clauses_permutation is None:
        clauses_permutation = list(range(M))
        random.shuffle(clauses_permutation)

    # load clauses
    out._clauses = [None] * M
    for (old, new) in enumerate(clauses_permutation):
        out._clauses[new] = tuple(substitution[lit]
                                  for lit in cnf._clauses[old])

    # return the formula
    assert out._check_coherence(force=True)
    return out



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Random CNF Formulas
"""

import itertools
import random

from cnfformula.cnf import CNF


def clause_satisfied(cls, assignments):
    """Test whether a clause is satisfied by all assignments

Test if clauses `cls` is satisfied by all assigment in the
list assignments.
"""
    for assignment in assignments:

        if any(var in assignment and polarity == assignment[var]
               for (polarity, var) in cls):
            continue

        else:
            return False

    return True


def sample_clauses(k, indices, m, planted_assignments):
    clauses = set()
    t = 0
    while len(clauses) < m and t < 10 * m:
        t += 1

        cls = tuple((random.choice([True, False]), 'x_{0}'.format(i))
                    for i in sorted(random.sample(indices, k)))

        if clause_satisfied(cls, planted_assignments):
            clauses.add(cls)

    if len(clauses) < m:
        return sample_clauses_dense(k, indices, m, planted_assignments)
    return clauses


def all_clauses(k, indices, planted_assignments):
    for domain in itertools.combinations(indices, k):
        for polarity in itertools.product([True, False], repeat=k):

            cls = list(zip(polarity, ('x_{0}'.format(i) for i in domain)))
            if clause_satisfied(cls, planted_assignments):
                yield cls


def sample_clauses_dense(k, indices, m, planted_assignments):
    return random.sample(list(all_clauses(k, indices, planted_assignments)), m)


def RandomKCNF(k, n, m, seed=None, planted_assignments=[]):
    """Build a random k-CNF

    Sample :math:`m` clauses over :math:`n` variables, each of width
    :math:`k`, uniformly at random. The sampling is done without
    repetition, meaning that whenever a randomly picked clause is
    already in the CNF, it is sampled again.

    Parameters
    ----------
    k : int
       width of each clause

    n : int
       number of variables to choose from. The resulting CNF object
       will contain n variables even if some are not mentioned in the clauses.

    m : int
       number of clauses to generate

    seed : hashable object
       seed of the random generator

    planted_assignments : iterable(dict), optional 
       a set of total/partial assigments such that all clauses in the formula 
       will be satisfied by all of them.

    Returns
    -------
    a CNF object

    Raises
    ------
    ValueError
        when some paramenter is negative, or when k>n.

    """
    if seed:
        random.seed(seed)

    if n < 0 or m < 0 or k < 0:
        raise ValueError("Parameters must be non-negatives.")

    if k > n:
        raise ValueError("Clauses cannot have more {} literals.".format(n))

    descr = "Random {}-CNF over {} variables and {} clauses".format(k, n, m)
    F = CNF(description=descr)

    indices = range(1, n + 1)
    for i in indices:
        F.add_variable('x_{0}'.format(i))
    try:
        for clause in sample_clauses(k, indices, m, planted_assignments):
            F.add_clause(list(clause), strict=True)
    except ValueError:
        raise ValueError(
            "There are fewer clauses available than the number requested")

    return F



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Formula Helpers for the Pitfall formula

Copyright (C) 2012-2020 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""

from cnfformula.families.pitfall import PitfallFormula

from .formula_helpers import FormulaHelper
from .cmdline import positive_int, positive_even_int

usage_string = """Pitfall Formula

    The Pitfall formula was designed to be specifically easy for
    Resolution and hard for common CDCL heuristics. The formula is
    unsatisfiable and consists of three parts: an easy formula, a hard
    formula, and a pitfall misleading the solver into working with the
    hard part.

    The hard part are several copies of an unsatisfiable Tseitin
    formula on a random regular graph. The pitfall part is made up of
    a few gadgets over (primarily) two sets of variables: pitfall
    variables, which point the solver towards the hard part after
    being assigned, and safety variables, which prevent the gadget
    from breaking even if a few other variables are assigned.

    For more details, see the corresponding paper (Marc Vinyals, AAAI
    2020).

positional arguments:
 {0} v,d,ny,nz,k

 where:
 v    --- number of vertices of the Tseitin graph
 d    --- degree of the Tseitin graph
 ny   --- number of pitfall variables
 nz   --- number of safety variables
 k    --- number of copies of the hard and pitfall parts; controls how
          easy the easy part is
"""

example_string = """example:
 {0} 45 4 30 5 8      --- parameters used in the paper
 """

class PitfallCmdHelper(FormulaHelper):
    name = 'pitfall'
    description = 'Pitfall formula'

    @staticmethod
    def setup_command_line(parser):

        parser.usage = usage_string.format(parser.prog)

        parser.add_argument('v',
                            type=positive_int,
                            help="number of vertices of the Tseitin graph")
        parser.add_argument('d',
                            type=positive_int,
                            help="degree of the Tseitin graph")
        parser.add_argument('ny',
                            type=positive_int,
                            help="number of pitfall variables")
        parser.add_argument('nz',
                            type=positive_int,
                            help="number of safety variables")
        parser.add_argument('k',
                            type=positive_even_int,
                            help="number of copies of the hard and pitfall parts")

    @staticmethod
    def build_cnf(args):

        return PitfallFormula(args.v, args.d, args.ny, args.nz, args.k)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Build and manipulate CNF formulas

The module `contains facilities to generate cnf formulas, in
order to be printed in DIMACS or LaTeX formats. Such formulas are
ready to be fed to sat solvers.

The module implements the `CNF` object, which is the main entry point
to the `cnfformula` library. 



Copyright (C) 2012-2020  Massimo Lauria <lauria.massimo@gmail.com>
https://github.com/MassimoLauria/cnfgen.git

"""

from itertools import product, islice
from itertools import combinations, combinations_with_replacement
from collections import Counter, OrderedDict
import re
from math import ceil, log

from cnfformula.info import info
from .graphs import bipartite_sets, neighbors


class CNF(object):
    """Propositional formulas in conjunctive normal form.

    A CNF  formula is a  sequence of  clauses, which are  sequences of
    literals. Each literal is either a variable or its negation.

    Use ``add_variable`` method to add a variable to the formula. Two
    variable with the same name are considered the same variable, add
    successive additions do not have any effect.

    Use ``add_clause`` to add new clauses to CNF. Clauses will be added
    multiple times in case of multiple insertion of the same clauses.

    For documentation purpose it is possible use have an additional
    comment header at the top of the formula, which will be
    *optionally* exported to LaTeX or dimacs.

    Implementation:  for efficiency reason clauses and variable can
    only be added, and not deleted. Furthermore order matters in
    the representation.

    Examples
    --------
    >>> c=CNF([ [(True,"x1"),(True,"x2"),(False,"x3")], \
              [(False,"x2"),(True,"x4")] ])
    >>> print( c.dimacs(export_header=False) )
    p cnf 4 2
    1 2 -3 0
    -2 4 0
    >>> c.add_clause( [(False,"x3"),(True,"x4"),(False,"x5")] )
    >>> print( c.dimacs(export_header=False))
    p cnf 5 3
    1 2 -3 0
    -2 4 0
    -3 4 -5 0
    """
    def __init__(self, clauses=None, description=None):
        """Propositional formulas in conjunctive normal form.

        Parameters
        ----------
        clauses : ordered list of clauses
            a clause with k literals list containing k pairs, each
            representing a literal (see `add_clause`). First element
            is the polarity and the second is the variable, which must
            be an hashable object.

            E.g. (not x3) or x4 or (not x2) is encoded as [(False,"x3"),(True,"x4"),False,"x2")]

        description: string, optional
            a description of the formula
        """

        self.header = OrderedDict()
        if description is not None:
            self.header['description'] = description

        self.header['generator'] = "{} ({})".format(info['project'],
                                                    info['version'])
        self.header['copyright'] = info['copyright']
        self.header['url'] = info['url']

        # Initial empty formula
        self._clauses = []

        # Variable indexes <--> Variable names correspondence
        # first variable is indexed with 1.
        self._index2name = [None]
        self._name2index = dict()
        self._name2descr = dict()

        # Internal coherence can be disrupted by some methods.  API
        # methods require it to be rechecked.
        self._coherent = True

        # Load the initial data into the CNF
        for c in clauses or []:
            self.add_clause(c)

    #
    # Implementation of some standard methods
    #

    def __iter__(self):
        """Iterates over all clauses of the CNF
        """
        for cls in self._clauses:
            assert self._coherent
            yield self._uncompress_clause(cls)

    def __str__(self):
        """String representation of the formula
        """
        assert self._coherent
        n = len(self._index2name) - 1
        m = len(self)
        text = "CNF with {} vars and {} clauses".format(n, m)
        if 'description' in self.header:
            text += " -- " + self.header['description']
        return text

    def __len__(self):
        """Number of clauses in the formula
        """
        return len(self._clauses)

    #
    # Internal implementation methods, use at your own risk!
    #

    def _uncompress_clause(self, clause):
        """(INTERNAL USE) Uncompress a clause from the numeric representation.

        Arguments:
        - `clause`: clause to be uncompressed

        >>> c=CNF()
        >>> c.add_clause([(True,"x"),(False,"y")])
        >>> print(c._uncompress_clause([-1,-2]))
        [(False, 'x'), (False, 'y')]
        """
        return [(l > 0, self._index2name[abs(l)]) for l in clause]

    def _compress_clause(self, clause):
        """Convert a clause to its numeric representation.

        For reason of efficiency, clauses are memorized as tuples of
        integers. Each integer correspond to a variable, with sign +1
        or -1 depending whether it represents a positive or negative
        literal. The correspondence between the numbers and the
        variables names depends on the formula itself

        Parameters
        ----------
        clause: list of pairs
            a clause, in the form of a list of literals, which are
            pairs (bool,string).

        Returns
        -------
        a tuple of int

        Examples
        --------
        >>> c=CNF()
        >>> c.add_clause([(True,"x"),(False,"y")])
        >>> print(c._compress_clause([(False, 'x'), (False, 'y')]))
        (-1, -2)

        """
        return tuple((1 if p else -1) * self._name2index[n] for p, n in clause)

    def _add_compressed_clauses(self, clauses):
        """(INTERNAL USE) Add to the CNF a list of compressed clauses.

        This method uses the internal compressed clause representation
        to add a large batch of data  into the CNF.  It does not check
        for internal  coherence conditions,  and it  does not  need to
        convert between  internal and external  clause representation,
        so it  is very fast.   When assertions  are tested, a  call to
        this method will  disable the standard API, since  the CNF can
        be in an inconsistent state.

        Whenever the high level API is used with an inconsisten state
        the code will fail some assertion.

        In particular it does not check if the indexes correspond to a
        variable in the formula.

        To test consistency and re-enable the API, please call method
        `CNF._check_coherence`.

        Arguments:
        - `clauses`: a sequence of compressed clauses.

        >>> c=CNF()

        We add the variables in advance, so that the internal status
        stays coherent.

        >>> c.add_variable("x")
        >>> c.add_variable("y")
        >>> c.add_variable("z")

        When we add some compressed clauses, we need to test the
        internal status of the object. If the test is positive, then
        the high level API is available again.

        >>> c._add_compressed_clauses([[-1,2,3],[-2,1],[1,-3]])
        >>> c._check_coherence()
        True
        >>> print(c.dimacs(export_header=False))
        p cnf 3 3
        -1 2 3 0
        -2 1 0
        1 -3 0

        If we call the internal API several times, we need to test the
        object only once.

        >>> c._add_compressed_clauses([[-2,-3]])
        >>> c._add_compressed_clauses([[-1, 2]])
        >>> c._check_coherence()
        True
        >>> print(c.dimacs(export_header=False))
        p cnf 3 5
        -1 2 3 0
        -2 1 0
        1 -3 0
        -2 -3 0
        -1 2 0
        """
        self._coherent = False
        self._clauses.extend(tuple(c) for c in clauses)

    def _check_coherence(self, force=False):
        """Check if the formula is internally consistent.

        Certain fast manipulation methods are not safe if used
        incorrectly, so the CNF object may be corrupted. This method
        tests if that was not the case.

        Arguments:
        - `force`: force check even if the formula claims coherence

        >>> c=CNF()
        >>> c.add_variable("x")
        >>> c.add_variable("y")

        We add clauses mentioning three variables, and the formula is
        not coherent.

        >>> c._add_compressed_clauses([(-1,2),(1,-2),(1,3)])
        >>> c._check_coherence()
        False

        We cannot use the API now

        >>> c.clauses()
        Traceback (most recent call last):
        AssertionError
        """
        if not force and self._coherent:
            return True

        varindex = self._name2index
        varnames = self._index2name

        # number of variables and clauses
        N = len(varindex.keys())

        # Consistency in the variable dictionary
        if N != len(varnames) - 1:
            return False

        for i in range(1, N + 1):
            if varindex[varnames[i]] != i:
                return False

        # Count clauses and check literal representation
        for clause in self._clauses:
            for literal in clause:
                if not 0 < abs(literal) <= N:
                    return False

        # formula passed all tests
        self._coherent = True
        return True

    #
    # High level API: build the CNF
    #

    def add_clause(self,
                   clause,
                   literal_repetitions=False,
                   opposite_literals=False,
                   auto_variables=True,
                   strict=False):
        """Add a clause to the CNF.

        E.g. (not x3) or x4 or (not x2) is encoded as
             [(False,u"x3"),(True,u"x4"),(False,u"x2")]

        All variable mentioned in the clause will be added to the list
        of variables  of the CNF,  in the  order of appearance  in the
        clauses, unless `auto_variable` is ``False``.
        
        Parameters
        ----------
        clause: list of (bool,str) 
            the clause to be added in the CNF

            A clause with k literals is a list with k pairs.
            First coords are the polarities, second coords are utf8
            encoded strings with variable names.

            Clause may contain repeated or opposite literal, but this
            behavior can be modified by the optional flags. 

            Clauses are added with repetition, i.e. if the same clause
            is added twice then it will occur twice in the
            formula too.

        literal_repetitions: bool, optional
            True if and only if the clause can have repeated literal.

            Useful for sanity check. If the flag is `False` and the
            clause contain two copies of the same literal, then
            `ValueError` is raised. (default: False)

        opposite_literals: bool, optional
            True if and only if the clause can have opposite literal.

            Useful for sanity check. If the flag is `False` and the
            clause contain two opposite literals, then `ValueError`
            is raised. (default: False)

        auto_variables: bool, optional
            If `True` the clause can contain new variables.

            New variables occurring in the clause will be added to the
            formula, unless the flag is `False`. In that case when
            a clause contains an unknow variables, `ValueError` is
            raised. (default: True)

        strict: bool, optional
            If `True` impose restrictions on the clause.

            Setting this to `True` is equivalent to set
            `literal_repetitions`, `opposite_literals`,
            `auto_variables` to `False`. In case of conflicting
            settings, the most restrictive one holds.
            (default: False)
        """
        assert self._coherent

        # A clause must be an immutable object
        try:
            hash(tuple(clause))
        except TypeError:
            raise TypeError("%s is not a well formatted clause" % clause)

        # Activate the most restrictive setting

        literal_repetitions = literal_repetitions and (not strict)
        opposite_literals = opposite_literals and (not strict)
        auto_variables = auto_variables and (not strict)

        # Check literal repetitions
        if (not literal_repetitions) and len(set(clause)) != len(clause):
            raise ValueError(
                "Forbidden repeated literals in clause {}".format(clause))

        # Check opposite literals
        if not opposite_literals:
            positive = set([v for (p, v) in clause if p])
            negative = set([v for (p, v) in clause if not p])
            if len(positive & negative) > 0:
                emsg = "{ " + ", ".join(positive & negative) + " }"
                raise ValueError(
                    "Following variable occur with opposite literals: {}".
                    format(emsg))

        # Add the compressed clause
        try:
            self._clauses.append(self._compress_clause(clause))
        except KeyError as error:
            if not auto_variables:
                raise ValueError(
                    "The clause contains unknown variable: {}".format(error))
            else:
                for _, var in clause:
                    self.add_variable(var)
                self._clauses.append(self._compress_clause(clause))

    def add_clause_unsafe(self, clause):
        """Add a clause without checking input

        This is logically equivalent to :py:meth:`CNF.add_clause`
        where `literal_repetition`, `opposite_literals` and
        `auto_variables` are ``True``, but it is faster because
        it does less checks on the input.

        Parameters
        ----------
        clause: list of (bool,str) 
            the clause to be added in the CNF

            A clause with k literals is a list with k pairs.
            First coords are the polarities, second coords are utf8
            encoded strings with variable names.

            Clauses are added with repetition, i.e. if the same clause
            is added twice then it will occur twice in the
            formula too.

        Raises
        ------
        KeyError
            if the clause contains a variable which was not added to the formula before.
        """
        self._clauses.append(self._compress_clause(clause))

    def add_variable(self, var, description=None):
        """Add a variable to the formula (if not already resent).

        The variable must be `hashable`. I.e. it must be usable as key
        in a dictionary.  It raises `TypeError` if the variable cannot
        be hashed.

        Parameters
        ----------
        var: hashable
             the variable name to be added/updated. It can be any
             hashable object (i.e. a dictionary key).
        description: str, optional
             an explanation/description/comment about the variable.
        """
        assert self._coherent
        try:
            if not var in self._name2index:
                # name correpsond to the last variable so far
                self._index2name.append(var)
                self._name2index[var] = len(self._index2name) - 1
        except TypeError:
            raise TypeError("%s is not a legal variable name" % var)

        # update description
        if description is not None:
            self._name2descr[var] = description

    #
    # High level API: read the CNF
    #
    def variables(self):
        """Returns (a copy of) the list of variable names.
        """
        assert self._coherent
        vars_iterator = iter(self._index2name)
        next(vars_iterator)
        return vars_iterator

    def clauses(self):
        """Return the list of clauses
        """
        assert self._coherent
        return self.__iter__()

    def dimacs(self, export_header=True, extra_text=None):
        """Produce the dimacs encoding of the formula

        The formula is rendered in the DIMACS format for CNF formulas,
        which is a particularly popular input format for SAT solvers [1]_.

        .. note:: By default the DIMACS output is *ascii* encoded,
                  with non-ascii characters replaced.

        Parameters
        ----------
        export_header : bool
            determines whether the formula header should be inserted as
            a comment in the DIMACS output.

        extra_text : str, optional
            Additional text attached to the header

        Returns
        -------
        string
            the string contains the Dimacs code


        Examples
        --------
        >>> c=CNF([[(False,"x_1"),(True,"x_2"),(False,"x_3")],\
                   [(False,"x_2"),(False,"x_4")], \
                   [(True,"x_2"),(True,"x_3"),(False,"x_4")]])
        >>> print(c.dimacs(export_header=False))
        p cnf 4 3
        -1 2 -3 0
        -2 -4 0
        2 3 -4 0

        >>> c=CNF()
        >>> print(c.dimacs(export_header=False))
        p cnf 0 0
        <BLANKLINE>

        References
        ----------
        .. [1] http://www.cs.ubc.ca/~hoos/SATLIB/Benchmarks/SAT/satformat.ps

        """
        from io import StringIO
        output = StringIO()
        self._dimacs_dump_clauses(output, export_header, extra_text)
        return output.getvalue()

    def _dimacs_dump_clauses(self,
                             output=None,
                             export_header=True,
                             extra_text=None):
        """Dump the dimacs encoding of the formula to the file-like output

        This is for internal use only. It produces the dimacs output
        of the clauses, and write then on the output buffer, which is
        tipically a StringIO.
        """
        assert self._coherent

        # Count the number of variables and clauses
        n = len(self._index2name) - 1
        m = len(self)

        # Produce header in ascii compatible format
        if export_header:
            # remove non ascii text
            for field in self.header:
                tmp = "c {}: {}\n".format(field, self.header[field])
                tmp = tmp.encode('ascii', errors='replace').decode('ascii')
                output.write(tmp)
            output.write("c\n")

            # remove non ascii text
            if extra_text is not None:
                ascii_extra = extra_text.encode('ascii', errors='replace')
                ascii_extra = ascii_extra.decode('ascii')
                for line in ascii_extra.split("\n"):
                    output.write(("c " + line).rstrip() + "\n")

        # Formula specification
        output.write("p cnf {0} {1}".format(n, m))

        if len(self._clauses) == 0:
            output.write("\n")  # this newline makes `lingeling` solver happy

        # Clauses
        for cls in self._clauses:
            output.write("\n" + " ".join([str(l) for l in cls + (0, )]))

    def latex(self, export_header=True, extra_text=None, full_document=False):
        """Output a LaTeX version of the CNF formula

        The CNF formula is translated into the LaTeX markup language
        [1]_, using the names of the variable literally. The formula
        is rendered in the ``align`` environment, with one clause per
        row. Negated literals are rendered using the
        ``\\neg`` command.

        The output string is ready to be included in a document, but
        it does not include neither a preamble nor is nested inside
        ``\\begin{document}`` ... ``\\end{document}``.

        .. note::  By default the LaTeX document in output is *UTF-8* encoded.

        Parameters
        ----------
        export_header : bool, optional
            determines whether the formula header should be inserted as
            a LaTeX comment in the output. By default is True. 

        extra_text : str, optional
            Additional text attached to the abstract.

        full_document : bool, optional
            rather than just output the formula, output a document 
            that contains it. False by default.

        Returns
        -------
        string
            the string contains the LaTeX code

        Examples
        --------
        >>> c=CNF([[(False,"x_1"),(True,"x_2"),(False,"x_3")],\
                   [(False,"x_2"),(False,"x_4")], \
                   [(True,"x_2"),(True,"x_3"),(False,"x_4")]])
        >>> print(c.latex(export_header=False))
        \\begin{align}
        &       \\left( {\\overline{x}_1} \\lor            {x_2} \\lor {\\overline{x}_3} \\right) \\\\
        & \\land \\left( {\\overline{x}_2} \\lor {\\overline{x}_4} \\right) \\\\
        & \\land \\left(            {x_2} \\lor            {x_3} \\lor {\\overline{x}_4} \\right)
        \\end{align}
        >>> c=CNF()
        >>> print(c.latex(export_header=False))
        \\begin{align}
           \\top
        \\end{align}

        References
        ----------
        .. [1] http://www.latex-project.org/
        """
        assert self._coherent

        clauses_per_page = 40

        latex_preamble = r"""%
\documentclass[10pt,a4paper]{article}
\usepackage[margin=1in]{geometry}
\usepackage{amsmath}
\usepackage{listings}
\usepackage[utf8]{inputenc}
"""

        from io import StringIO
        output = StringIO()

        # formula header as a LaTeX comment
        if export_header:
            for field in self.header:
                tmp = "% {}: {}\n".format(field, self.header[field])
                tmp = tmp.encode('ascii', errors='replace').decode('ascii')
                output.write(tmp)
            output.write("%\n")

        # document opening
        if full_document:
            output.write(latex_preamble)
            output.write("\\begin{document}\n")
            title = self.header['description']
            title = title.replace("_", "\\_")
            output.write("\\title{{{}}}\n".format(title))
            output.write("\\author{CNFgen formula generator}\n")
            output.write("\\maketitle\n")
            output.write("\\noindent\\textbf{Formula header:}\n")
            output.write("\\begin{lstlisting}[breaklines]\n")
            for field in self.header:
                tmp = "{}: {}\n".format(field, self.header[field])
                tmp = tmp.encode('ascii', errors='replace').decode('ascii')
                output.write(tmp)
            output.write("\\end{lstlisting}\n")
            output.write("\\bigskip\n")

        if extra_text is not None and full_document:
            output.write(extra_text)

        def map_literals(l):
            """Map literals to LaTeX string"""
            assert l != 0
            if l > 0:
                return "           {" + str(self._index2name[l]) + "}"
            else:
                name = self._index2name[-l]
                split_point = name.find("_")
                if split_point < 1:
                    return "  \\overline{" + name + "}"
                else:
                    return "{\\overline{" + name[:split_point] + "}" + name[
                        split_point:] + "}"

        def write_clause(cls, first, full_document):
            """Write the clause in LaTeX."""
            output.write("\n&" if first else " \\\\\n&")
            output.write("       " if full_document or first else " \\land ")

            # build the latex clause
            if len(cls) == 0:
                output.write("\\square")
            elif full_document:
                output.write(" \\lor ".join(map_literals(l) for l in cls))
            else:
                output.write("\\left( " + \
                             " \\lor ".join(map_literals(l) for l in cls) + \
                             " \\right)")

        # Output the clauses
        clauses_number = len(self._clauses)
        if full_document:
            output.write("\\noindent\\textbf{{CNF with {} variables and and {} clauses:}}\n".\
                         format(len(self._name2index),clauses_number))

        output.write("\\begin{align}")

        if clauses_number == 0:
            output.write("\n   \\top")
        else:
            for i, clause in enumerate(self._clauses):
                if i % clauses_per_page == 0 and i != 0 and full_document:
                    output.write("\n\\end{align}\\pagebreak")
                    output.write("\n\\begin{align}")
                    write_clause(clause, True, full_document)
                else:
                    write_clause(clause, i == 0, full_document)

        output.write("\n\\end{align}")

        # document closing
        if full_document:
            output.write("\n\\end{document}")

        return output.getvalue()

    def is_satisfiable(self, cmd=None, sameas=None, verbose=0):
        """Determines whether a CNF is satisfiable or not.

        The formula is passed to a SAT solver, according to the
        optional command line ``cmd``. If no command line is
        specified, the known solvers are tried in succession until one
        is found.

        It is possible to use any drop-in replacement for these
        solvers, but in this case more information is needed on how to
        communicate with the solver. In particular ``minisat`` does not
        respect the standard DIMACS I/O conventions, and that holds
        also for ``glucose`` which is a drop-in replacement of
        ``minisat``.

        For the supported solver we can pick the right interface, but
        for other solvers it is impossible to guess. Nevertheless it
        is possible to indicate which interface to use, or more
        specifically which known solver interface to mimic.

        >>> F.is_satisfiable(cmd='minisat-style-solver',sameas='minisat')  # doctest: +SKIP
        >>> F.is_satisfiable(cmd='dimacs-style-solver',sameas='lingeling') # doctest: +SKIP

        Parameters
        ----------
        cmd : string,optional
            the actual command line used to invoke the SAT solver

        sameas : string, optional
            use the interface of one of the supported solvers. Useful
            when the solver used in the command line is not supported.

        verbose: int
            0 or less means no output. 1 shows the command line actually
            run. 2 outputs the solver output. (default: 0)


        Examples
        --------
        >>> F.is_satisfiable()                                              # doctest: +SKIP
        >>> F.is_satisfiable(cmd='minisat -no-pre')                         # doctest: +SKIP
        >>> F.is_satisfiable(cmd='glucose -pre')                            # doctest: +SKIP
        >>> F.is_satisfiable(cmd='lingeling --plain')                       # doctest: +SKIP
        >>> F.is_satisfiable(cmd='sat4j')                                   # doctest: +SKIP
        >>> F.is_satisfiable(cmd='my-hacked-minisat -pre',sameas='minisat') # doctest: +SKIP
        >>> F.is_satisfiable(cmd='patched-lingeling',sameas='lingeling')    # doctest: +SKIP

        Returns
        -------
        (boolean,assignment or None)
            A pair (answer,witness) where answer is either True when
            F is satisfiable, or False otherwise. If F is satisfiable
            the witness is a satisfiable assignment in form of
            a dictionary, otherwise it is None.

        Raises
        ------
        RuntimeError
           if it is not possible to correctly invoke the solver needed.

        ValueError
           if `sameas` is set and is not the name of a supported solver.

        TypeError
           if F is not a CNF object.

        See Also
        --------
        cnfformula.utils.solver.is_satisfiable : implementation independent of CNF object.
        cnfformula.utils.solver.supported_satsolvers : the SAT solver recognized by `cnfformula`.

        """
        from .utils import solver
        return solver.is_satisfiable(self,
                                     cmd=cmd,
                                     sameas=sameas,
                                     verbose=verbose)

    ###
    ### Various utility function for CNFs
    ###
    @classmethod
    def parity_constraint(cls, variables, constant):
        """Output the CNF encoding of a parity constraint
        
        E.g. X1 + X2 + X3 = 1 (mod 2) is encoded as
        
        ( X1 v  X2 v  X3)
        (~X1 v ~X2 v  X3)
        (~X1 v  X2 v ~X3)
        ( X1 v ~X2 v ~X3)

        Parameters
        ----------
        variables : array-like
            variables involved in the constraint
        constant : {0,1}
            the constant of the linear equation

        Returns
        -------
        a list of clauses

        Examples
        --------
        >>> list(CNF.parity_constraint(['a','b'],1))
        [[(True, 'a'), (True, 'b')], [(False, 'a'), (False, 'b')]]
        >>> list(CNF.parity_constraint(['a','b'],0))
        [[(True, 'a'), (False, 'b')], [(False, 'a'), (True, 'b')]]
        >>> list(CNF.parity_constraint(['a'],0))
        [[(False, 'a')]]
        """
        domains = tuple([((True, var), (False, var)) for var in variables])
        clauses = []
        for c in product(*domains):
            # Save only the clauses with the right polarity
            parity = sum(1 - l[0] for l in c) % 2
            if parity != constant:
                yield list(c)

    @classmethod
    def _inequality_constraint_builder(cls, variables, k, greater=False):
        """Builder for inequality constraint
     
        This is a generic builder used to build all the inequality
        constraints. By default it build a "stricly less that", and if
        ``greater`` is True it builds a "strictly greater than".
        """
        polarity = greater
        if greater:
            k = len(variables) - k

        if k > len(variables):
            return
        elif k < 0:
            yield []
            return

        for tpl in combinations(variables, k):
            yield [(polarity, v) for v in tpl]

    @classmethod
    def less_than_constraint(cls, variables, upperbound):
        """Clauses encoding a \"strictly less than\" constraint
     
        E.g. X1 + X2 + X3 + X4 < 3
     
        (~X1 v ~X2 v ~X3)
        (~X1 v ~X2 v ~X4)
        (~X1 v ~X3 v ~X4)
        (~X2 v ~X3 v ~X4)
     
        Parameters
        ----------
        variables : list of variables
           variables in the constraint
        upperbound: int
           upper bound of the constraint
     
        Returns
        -------
            a list of clauses
     
        Examples
        --------
        >>> list(CNF.less_than_constraint(['a','b','c'],2))
        [[(False, 'a'), (False, 'b')], [(False, 'a'), (False, 'c')], [(False, 'b'), (False, 'c')]]
        >>> list(CNF.less_than_constraint(['a'],1))
        [[(False, 'a')]]
        >>> list(CNF.less_than_constraint(['a','b','c'],-1))
        [[]]
        >>> list(CNF.less_than_constraint(['a','b','c'],10))
        []
        """
        return cls._inequality_constraint_builder(variables,
                                                  upperbound,
                                                  greater=False)

    @classmethod
    def less_or_equal_constraint(cls, variables, upperbound):
        """Clauses encoding a \"less than or equal to\" constraint
     
        E.g. X1 + X2 + X3 + X4 <= 2
     
        (~X1 v ~X2 v ~X3)
        (~X1 v ~X2 v ~X4)
        (~X1 v ~X3 v ~X4)
        (~X2 v ~X3 v ~X4)
     
        Parameters
        ----------
        variables : list of variables
           variables in the constraint
        upperbound: int
           upper bound of the constraint
     
        Returns
        -------
            a list of clauses
     
        Examples
        --------
        >>> list(CNF.less_than_constraint(['a','b','c'],3)) == list(CNF.less_or_equal_constraint(['a','b','c'],2))
        True
        >>> list(CNF.less_or_equal_constraint(['a','b','c'],1))
        [[(False, 'a'), (False, 'b')], [(False, 'a'), (False, 'c')], [(False, 'b'), (False, 'c')]]
        >>> list(CNF.less_or_equal_constraint(['a','b'],0))
        [[(False, 'a')], [(False, 'b')]]
        >>> list(CNF.less_or_equal_constraint(['a','b','c'],-1))
        [[]]
        >>> list(CNF.less_or_equal_constraint(['a','b','c'],10))
        []
        """
        return cls._inequality_constraint_builder(variables,
                                                  upperbound + 1,
                                                  greater=False)

    @classmethod
    def greater_than_constraint(cls, variables, lowerbound):
        """Clauses encoding a \"strictly greater than\" constraint
     
        E.g. X1 + X2 + X3 + X4 > 2
     
        (X1 v X2 v X3)
        (X1 v X2 v X4)
        (X1 v X3 v X4)
        (X2 v X3 v X4)
     
        Parameters
        ----------
        variables : list of variables
           variables in the constraint
        lowerbound: int
           lower bound of the constraint
     
        Returns
        -------
            a list of clauses
     
        Examples
        --------
        >>> list(CNF.greater_than_constraint(['a','b','c'],2))
        [[(True, 'a')], [(True, 'b')], [(True, 'c')]]
        >>> list(CNF.greater_than_constraint(['a'],0))
        [[(True, 'a')]]
        >>> list(CNF.greater_than_constraint(['a','b','c'],-1))
        []
        >>> list(CNF.greater_than_constraint(['a','b','c'],3))
        [[]]
        """
        return cls._inequality_constraint_builder(variables,
                                                  lowerbound,
                                                  greater=True)

    @classmethod
    def greater_or_equal_constraint(cls, variables, lowerbound):
        """Clauses encoding a \"greater than or equal to\" constraint
     
        E.g. X1 + X2 + X3 + X4 > 1
     
        (X1 v X2 v X3)
        (X1 v X2 v X4)
        (X1 v X3 v X4)
        (X2 v X3 v X4)
     
        Parameters
        ----------
        variables : list of variables
           variables in the constraint
        lowerbound: int
           lower bound of the constraint
     
        Returns
        -------
            a list of clauses
     
        Examples
        --------
        >>> tuple(CNF.greater_than_constraint(['a','b','c'],1)) == tuple(CNF.greater_or_equal_constraint(['a','b','c'],2))
        True
        >>> list(CNF.greater_or_equal_constraint(['a','b','c'],3))
        [[(True, 'a')], [(True, 'b')], [(True, 'c')]]
        >>> list(CNF.greater_or_equal_constraint(['a'],0))
        []
        >>> list(CNF.greater_or_equal_constraint(['a','b','c'],4))
        [[]]
        """
        return cls._inequality_constraint_builder(variables,
                                                  lowerbound - 1,
                                                  greater=True)

    @classmethod
    def equal_to_constraint(cls, variables, value):
        """Clauses encoding a \"equal to\" constraint
     
        E.g. X1 + X2 + X3 + X4 = 1
     
        (X1 v X2 v X3 v X4)
        (~X1 v ~X2)
        (~X1 v ~X3)
        (~X1 v ~X4)
        (~X2 v ~X3)
        (~X2 v ~X4)
        (~X3 v ~X4)
     
        Parameters
        ----------
        variables : list of variables
           variables in the constraint
        value: int
           target values
     
        Returns
        -------
            a list of clauses
        """
        for c in cls.less_or_equal_constraint(variables, value):
            yield c
        for c in cls.greater_or_equal_constraint(variables, value):
            yield c

    @classmethod
    def loose_majority_constraint(cls, variables):
        """Clauses encoding a \"at least half\" constraint
     
        Parameters
        ----------
        variables : list of variables
           variables in the constraint
     
        Returns
        -------
            a list of clauses
        """
        threshold = int((len(variables) + 1) / 2)
        return cls.greater_or_equal_constraint(variables, threshold)

    @classmethod
    def loose_minority_constraint(cls, variables):
        """Clauses encoding a \"at most half\" constraint
     
        Parameters
        ----------
        variables : list of variables
           variables in the constraint
     
        Returns
        -------
            a list of clauses
        """
        threshold = int(len(variables) / 2)
        return cls.less_or_equal_constraint(variables, threshold)

    @classmethod
    def exactly_half_ceil(cls, variables):
        """Clauses encoding a \"exactly half\" constraint (rounded up)
     
        Parameters
        ----------
        variables : list of variables
           variables in the constraint
     
        Returns
        -------
            a list of clauses
        """
        threshold = int((len(variables) + 1) / 2)
        return cls.equal_to_constraint(variables, threshold)

    @classmethod
    def exactly_half_floor(cls, variables):
        """Clauses encoding a \"exactly half\" constraint (rounded down)
     
        Parameters
        ----------
        variables : list of variables
           variables in the constraint
     
        Returns
        -------
            a list of clauses
        """
        threshold = int(len(variables) / 2)
        return cls.equal_to_constraint(variables, threshold)

    class unary_mapping(object):
        """Unary CNF representation of a mapping between two sets."""

        Domain = None
        Range = None

        Pattern = None

        Complete = False
        Functional = False
        Surjective = False
        Injective = False

        NonDecreasing = False

        @staticmethod
        def var_name(i, b):
            return "X_{{{0},{1}}}".format(i, b)

        def __init__(self, D, R, **kwargs):
            r"""Generator for the clauses of a mapping between to sets

            This generates of the constraints on variables :math:`v(i,j)`
            where :math:`i \in D` and :math:`j in R`, so that they
            represent a mapping (or a relation) between the two sets,
            expressed in unary (i.e. :math:`v(i,j)` expresses whether
            :math:`i` is mapped to :math:`j` or not).
            
            Parameters
            ----------
            D : iterable
                the domain of the mapping

            R : iterable
                the range of the mapping

            sparsity_pattern : bipartite_graph, optional
                each element of the domain is allowed to be mapped
                only into certain range elements. The graph represents
                which range elements are compatible with a specific
                domain element.

            var_name: function, optional 
                given :math:`i` and :math`j` the function must produce the
                name of variable :math`v(i,j)`
     
            complete: bool, optional
                every element of :math:`D` must have an image (default: true)
     
            functional: bool, optional
                every element of :math:`D` must have at most one image (default: false)
     
            surjective: bool, optional
                every element of :math:`R` must have a pre-image (default: false)
     
            injective: bool, optional
                every element of :math:`R` must have at most one pre-image (default: false)
     
            nondecreasing: bool, optional
                the mapping is going to be non decresing, with respect to
                the order of domain and range (default: false)
                
            """
            self.Domain = list(D)
            self.Range = list(R)

            # optional parameters of the mapping
            self.Complete = kwargs.pop('complete', True)
            self.Functional = kwargs.pop('functional', False)
            self.Surjective = kwargs.pop('surjective', False)
            self.Injective = kwargs.pop('injective', False)

            self.NonDecreasing = kwargs.pop('nondecreasing', False)

            # variable name scheme
            self.var_name = kwargs.pop('var_name', self.var_name)

            # use a bipartite graph scheme for the mapping?
            self.Pattern = kwargs.pop('sparsity_pattern', None)

            if kwargs:
                raise TypeError('Unexpected **kwargs: %r' % kwargs)

            self.DomainToVertex = {}
            self.VertexToDomain = {}
            self.RangeToVertex = {}
            self.VertexToRange = {}

            self.RankRange = {}
            self.RankDomain = {}

            if self.Pattern is not None:

                gL, gR = bipartite_sets(self.Pattern)

                if len(gL) != len(self.Domain):
                    raise ValueError(
                        "Domain and the left side of the pattern differ in size"
                    )

                if len(gR) != len(self.Range):
                    raise ValueError(
                        "Range and the right side of the pattern differ in size"
                    )
            else:
                gL = self.Domain
                gR = self.Range

            for (d, v) in zip(self.Domain, gL):
                self.DomainToVertex[d] = v
                self.VertexToDomain[v] = d

            for (r, v) in zip(self.Range, gR):
                self.RangeToVertex[r] = v
                self.VertexToRange[v] = r

            for i, d in enumerate(self.Domain, start=1):
                self.RankDomain[d] = i

            for i, r in enumerate(self.Range, start=1):
                self.RankRange[r] = i

        def domain(self):
            return self.Domain

        def range(self):
            return self.Range

        def images(self, d):
            if self.Pattern is None:
                return self.Range
            else:
                v = self.DomainToVertex[d]
                return [
                    self.VertexToRange[u] for u in neighbors(self.Pattern, v)
                ]

        def counterimages(self, r):
            if self.Pattern is None:
                return self.Domain
            else:
                v = self.RangeToVertex[r]
                return [
                    self.VertexToDomain[u] for u in neighbors(self.Pattern, v)
                ]

        def variables(self):
            for d in self.Domain:
                for r in self.images(d):
                    yield self.var_name(d, r)

        def clauses(self):

            # Completeness axioms
            if self.Complete:
                for d in self.Domain:
                    for c in CNF.greater_or_equal_constraint(
                        [self.var_name(d, r) for r in self.images(d)], 1):
                        yield c

            # Surjectivity axioms
            if self.Surjective:
                for r in self.Range:
                    for c in CNF.greater_or_equal_constraint(
                        [self.var_name(d, r) for d in self.counterimages(r)],
                            1):
                        yield c

            # Injectivity axioms
            if self.Injective:
                for r in self.Range:
                    for c in CNF.less_or_equal_constraint(
                        [self.var_name(d, r) for d in self.counterimages(r)],
                            1):
                        yield c

            # Functionality axioms
            if self.Functional:
                for d in self.Domain:
                    for c in CNF.less_or_equal_constraint(
                        [self.var_name(d, r) for r in self.images(d)], 1):
                        yield c

            # Mapping is monotone non-decreasing
            if self.NonDecreasing:

                for (a, b) in combinations(self.Domain, 2):
                    for (i, j) in product(self.images(a), self.images(b)):

                        if self.RankRange[i] > self.RankRange[j]:
                            yield [(False, self.var_name(a, i)),
                                   (False, self.var_name(b, j))]

    class binary_mapping(object):
        """Binary CNF representation of a mapping between two sets."""

        Domain = None
        Range = None
        Bits = None

        Injective = False
        NonDecreasing = False

        @staticmethod
        def var_name(i, b):
            return "Y_{{{0},{1}}}".format(i, b)

        def variables(self):
            for v, b in product(self.Domain, range(0, self.Bits)):
                yield self.var_name(v, b)

        def __init__(self, D, R, **kwargs):
            r"""Generator for the clauses of a binary mapping between D and :math:`R`
            
            This generates of the constraints on variables
            :math:`v(i,0)...v(i,k-1)` where :math:`i \in D` and
            :math:`v(i,0)...v(i,k-1)` is a binary of :math:`k` bits, so
            that the first :math:`|R|` string in :math:`{0,1}^k` (in
            lexicographic order) encode the elements of :math:`R`. 

            Parameters
            ----------
            D : iterable
                the domain of the mapping
            
            R : iterable
                the length of the bit strings

            var_name: function, optional
                given :math:`i` and :math`b` the function must produce the
                name of variable :math`v(i,b)`

            injective: bool, optional
                every bitstring must have at most one pre-image (default: false)
             
            nondecreasing: bool, optional
                the mapping must be non decreasing (default: false)

            """
            self.Domain = list(D)
            self.Range = list(R)
            self.Bits = int(ceil(log(len(R), 2)))

            # optional parameters of the mapping
            self.Injective = kwargs.pop('injective', False)
            self.NonDecreasing = kwargs.pop('nondecreasing', False)
            # variable name scheme
            self.var_name = kwargs.pop('var_name', self.var_name)

            if kwargs:
                raise TypeError('Unexpected **kwargs: %r' % kwargs)

            self.ImageToBits = {}
            self.BitsToImage = {}

            for i, bs in islice(enumerate(product([0, 1], repeat=self.Bits)),
                                len(self.Range)):

                self.ImageToBits[self.Range[i]] = bs
                self.BitsToImage[bs] = self.Range[i]

        def image_to_bitstring(self, im):
            return self.ImageToBits[im]

        def bitstring_to_image(self, bs):
            return self.BitsToImage[bs]

        def forbid_bitstring(self, i, bs):
            """Generates a clause that exclude 'i -> bs' mapping """
            return [(bs[b] == 0, self.var_name(i, self.Bits - 1 - b))
                    for b in range(self.Bits)]

        def forbid_image(self, i, j):
            """Generates a clause that exclude 'i -> j' mapping """
            return self.forbid_bitstring(i, self.ImageToBits[j])

        def clauses(self):

            # Avoid strings that do not correspond to elements from the range
            for i, bs in product(
                    self.Domain,
                    islice(product([0, 1], repeat=self.Bits), len(self.Range),
                           None)):
                yield self.forbid_bitstring(i, bs)

            # Injectivity
            if self.Injective:
                for j in self.Range:
                    for i1, i2 in combinations(self.Domain, 2):
                        yield self.forbid_image(i1, j) + self.forbid_image(
                            i2, j)

            # Enforce Non Decreasing Mapping
            if self.NonDecreasing:
                pairs_of_maps = product(combinations(self.Domain, 2),
                                        combinations(self.Range, 2))

                for (i1, i2), (j1, j2) in pairs_of_maps:
                    yield self.forbid_image(i1, j2) + self.forbid_image(i2, j1)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of some Ordering principle helpers

Copyright (C) 2012, 2013, 2014, 2015, 2016, 2019 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""

from cnfformula import OrderingPrinciple
from cnfformula import GraphOrderingPrinciple

from .graph_cmdline import SimpleGraphHelper
from .formula_helpers import FormulaHelper


class OPCmdHelper(FormulaHelper):
    """Command line helper for Ordering principle formulas
    """
    name='op'
    description='ordering principle'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for Ordering principle formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('N',metavar='<N>',type=int,help="domain size")
        g=parser.add_mutually_exclusive_group()
        g.add_argument('--total','-t',default=False,action='store_true',help="assume a total order")
        g.add_argument('--smart','-s',default=False,action='store_true',help="encode 'x<y' and 'x>y' in a single variable (implies totality)")
        g.add_argument('--knuth2', action='store_const', dest='knuth',const=2,
                       help="transitivity axioms: \"(i<j)(j<k)->(i,k)\" only for j>i,k")
        g.add_argument('--knuth3', action='store_const', dest='knuth',const=3,
                       help="transitivity axioms: \"(i<j)(j<k)->(i,k)\" only for k>i,j")
        parser.add_argument('--plant','-p',default=False,action='store_true',help="allow a minimum element")

    @staticmethod
    def build_cnf(args):
        """Build an Ordering principle formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        return OrderingPrinciple(args.N,args.total,args.smart,args.plant,args.knuth)


class GOPCmdHelper(FormulaHelper):
    """Command line helper for Graph Ordering principle formulas
    """
    name='gop'
    description='graph ordering principle'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for Graph ordering principle formula

        Arguments:
        - `parser`: parser to load with options.
        """
        g=parser.add_mutually_exclusive_group()
        g.add_argument('--total','-t',default=False,action='store_true',help="assume a total order")
        g.add_argument('--smart','-s',default=False,action='store_true',help="encode 'x<y' and 'x>y' in a single variable (implies totality)")
        g.add_argument('--knuth2', action='store_const', dest='knuth',const=2,
                       help="transitivity axioms: \"(i<j)(j<k)->(i,k)\" only for j>i,k")
        g.add_argument('--knuth3', action='store_const', dest='knuth',const=3,
                       help="transitivity axioms: \"(i<j)(j<k)->(i,k)\" only for k>i,j")
        parser.add_argument('--plant','-p',default=False,action='store_true',help="allow a minimum element")
        SimpleGraphHelper.setup_command_line(parser)


    @staticmethod
    def build_cnf(args):
        """Build a Graph ordering principle formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        G = SimpleGraphHelper.obtain_graph(args)
        return GraphOrderingPrinciple(G,args.total,args.smart,args.plant,args.knuth)





#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of the clique-coloring formula 
"""

from cnfformula.cnf import CNF
from itertools import combinations, permutations


def CliqueColoring(n, k, c):
    r"""Clique-coloring CNF formula 

    The formula claims that a graph :math:`G` with :math:`n` vertices
    simultaneously contains a clique of size :math:`k` and a coloring
    of size :math:`c`.

    If :math:`k = c + 1` then the formula is clearly unsatisfiable,
    and it is the only known example of a formula hard for cutting
    planes proof system. [1]_

    Variables :math:`e_{u,v}` to encode the edges of the graph.
    
    Variables :math:`q_{i,v}` encode a function from :math:`[k]` to
    :math:`[n]` that represents a clique.
    
    Variables :math:`r_{v,\ell}` encode a function from :math:`[n]` to
    :math:`[c]` that represents a coloring.
     
    Parameters
    ----------
    n : number of vertices in the graph
    k : size of the clique
    c : size of the coloring

    Returns
    -------
    A CNF object

    References
    ----------
    .. [1] Pavel Pudlak.
           Lower bounds for resolution and cutting plane proofs and
           monotone computations.
           Journal of Symbolic Logic (1997)

    """
    def E(u, v):
        "Name of an edge variable"
        return 'e_{{{0},{1}}}'.format(min(u, v), max(u, v))

    def Q(i, v):
        "Name of an edge variable"
        return 'q_{{{0},{1}}}'.format(i, v)

    def R(v, ell):
        "Name of an coloring variable"
        return 'r_{{{0},{1}}}'.format(v, ell)

    description = "There is a graph of {0} vertices with a {1}-clique and a {2}-coloring".format(
        n, k, c)
    formula = CNF(description=description)
    # Edge variables
    for u in range(1, n + 1):
        for v in range(u + 1, n + 1):
            formula.add_variable(E(u, v))
    # Clique encoding variables
    for i in range(1, k + 1):
        for v in range(1, n + 1):
            formula.add_variable(Q(i, v))
    # Coloring encoding variables
    for v in range(1, n + 1):
        for ell in range(1, c + 1):
            formula.add_variable(R(v, ell))

    # some vertex is i'th member of clique
    for k in range(1, k + 1):
        for cl in CNF.equal_to_constraint([Q(k, v) for v in range(1, n + 1)],
                                          1):
            formula.add_clause(cl, strict=True)

    # clique members are connected by edges
    for v in range(1, n + 1):
        for i, j in combinations(range(1, k + 1), 2):
            formula.add_clause([(False, Q(i, v)), (False, Q(j, v))],
                               strict=True)
    for u, v in combinations(range(1, n + 1), 2):
        for i, j in permutations(range(1, k + 1), 2):
            formula.add_clause([(True, E(u, v)), (False, Q(i, u)),
                                (False, Q(j, v))],
                               strict=True)

    # every vertex v has exactly one colour
    for v in range(1, n + 1):
        for cl in CNF.equal_to_constraint(
            [R(v, ell) for ell in range(1, c + 1)], 1):
            formula.add_clause(cl, strict=True)

    # neighbours have distinct colours
    for u, v in combinations(range(1, n + 1), 2):
        for ell in range(1, c + 1):
            formula.add_clause([(False, E(u, v)), (False, R(u, ell)),
                                (False, R(v, ell))],
                               strict=True)
    return formula



#!/usr/bin/env python
"""
Setup script for the CNFgen package
"""
from os import path
from setuptools import setup, find_packages

from cnfformula.info import info

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'PyPI.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=info['project'],
    version=info['version'],
    description=info['description'],
    author=info['author'],
    author_email=info['author_email'],
    url=info['url'],
    license=info['license'],
    packages=find_packages(".", exclude=["tests"]),
    entry_points={
        'console_scripts':
        ['cnfgen=cnfgen.main:main', 'cnfshuffle=cnfgen.cnfshuffle:main'],
    },
    install_requires=['networkx>=2.0', 'pydot>=1.2.3'],
    python_requires='>=3.4',

    # Package long description in Markdown
    long_description=long_description,
    long_description_content_type='text/markdown')



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of some graph formulas helpers

Copyright (C) 2012, 2013, 2014, 2015, 2016, 2019, 2020 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""

from cnfformula import GraphColoringFormula
from cnfformula import EvenColoringFormula

from cnfformula import DominatingSet

from cnfformula import GraphIsomorphism
from cnfformula import GraphAutomorphism

from cnfformula import CliqueFormula
from cnfformula import BinaryCliqueFormula
from cnfformula import SubgraphFormula
from cnfformula import RamseyWitnessFormula

from .graph_cmdline import SimpleGraphHelper
from .formula_helpers import FormulaHelper


class KColorCmdHelper(FormulaHelper):
    """Command line helper for k-color formula
    """
    name = 'kcolor'
    description = 'k-colorability formula'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for k-color formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('k',
                            metavar='<k>',
                            type=int,
                            action='store',
                            help="number of available colors")
        SimpleGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        """Build a k-colorability formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        G = SimpleGraphHelper.obtain_graph(args)
        return GraphColoringFormula(G, list(range(1, args.k + 1)))


class ECCmdHelper(FormulaHelper):
    name = 'ec'
    description = 'even coloring formulas'

    @staticmethod
    def setup_command_line(parser):
        SimpleGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        G = SimpleGraphHelper.obtain_graph(args)
        return EvenColoringFormula(G)


class DominatingSetCmdHelper(FormulaHelper):
    """Command line helper for k-dominating set
    """
    name = 'domset'
    description = 'k-Dominating set'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for dominating set formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('d',
                            metavar='<d>',
                            type=int,
                            action='store',
                            help="size of the dominating set")
        parser.add_argument('--alternative',
                            '-a',
                            action='store_true',
                            default=False,
                            help="produce the provably hard version")
        SimpleGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        """Build the k-dominating set formula

        Arguments:
        - `args`: command line options
        """
        G = SimpleGraphHelper.obtain_graph(args)
        return DominatingSet(G, args.d, alternative=args.alternative)


class GAutoCmdHelper(FormulaHelper):
    """Command line helper for Graph Automorphism formula
    """
    name = 'gauto'
    description = 'graph automorphism formula'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for graph automorphism formula

        Arguments:
        - `parser`: parser to load with options.
        """
        SimpleGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        """Build a graph automorphism formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        G = SimpleGraphHelper.obtain_graph(args)
        return GraphAutomorphism(G)


class GIsoCmdHelper(FormulaHelper):
    """Command line helper for Graph Isomorphism formula
    """
    name = 'giso'
    description = 'graph isomorphism formula'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for graph isomorphism formula

        Arguments:
        - `parser`: parser to load with options.
        """
        SimpleGraphHelper.setup_command_line(parser, suffix="1", required=True)
        SimpleGraphHelper.setup_command_line(parser, suffix="2", required=True)

    @staticmethod
    def build_cnf(args):
        """Build a graph automorphism formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        G1 = SimpleGraphHelper.obtain_graph(args, suffix="1")
        G2 = SimpleGraphHelper.obtain_graph(args, suffix="2")
        return GraphIsomorphism(G1, G2)


class KCliqueCmdHelper(FormulaHelper):
    """Command line helper for k-clique formula
    """
    name = 'kclique'
    description = 'k clique formula'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for k-clique formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('k',
                            metavar='<k>',
                            type=int,
                            action='store',
                            help="size of the clique to be found")
        SimpleGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        """Build a k-clique formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        G = SimpleGraphHelper.obtain_graph(args)
        return CliqueFormula(G, args.k)


class BinaryKCliqueCmdHelper(FormulaHelper):
    """Command line helper for k-clique formula
    """
    name = 'kcliquebin'
    description = 'Binary k clique formula'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for k-clique formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('k',
                            metavar='<k>',
                            type=int,
                            action='store',
                            help="size of the clique to be found")
        SimpleGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        """Build a k-clique formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        G = SimpleGraphHelper.obtain_graph(args)
        return BinaryCliqueFormula(G, args.k)


class RWCmdHelper(FormulaHelper):
    """Command line helper for ramsey graph formula
    """
    name = 'ramlb'
    description = 'unsat if G witnesses that r(k,s)>|V(G)| (i.e. G has not k-clique nor s-stable)'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for ramsey witness formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('k',
                            metavar='<k>',
                            type=int,
                            action='store',
                            help="size of the clique to be found")
        parser.add_argument('s',
                            metavar='<s>',
                            type=int,
                            action='store',
                            help="size of the stable to be found")
        SimpleGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        """Build a formula to check that a graph is a ramsey number lower bound

        Arguments:
        - `args`: command line options
        """
        G = SimpleGraphHelper.obtain_graph(args)
        return RamseyWitnessFormula(G, args.k, args.s)


class SubGraphCmdHelper(FormulaHelper):
    """Command line helper for Graph Isomorphism formula
    """
    name = 'subgraph'
    description = 'subgraph formula'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for graph isomorphism formula

        Arguments:
        - `parser`: parser to load with options.
        """
        SimpleGraphHelper.setup_command_line(parser, suffix="", required=True)
        SimpleGraphHelper.setup_command_line(parser, suffix="T", required=True)

    @staticmethod
    def build_cnf(args):
        """Build a graph automorphism formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        G = SimpleGraphHelper.obtain_graph(args, suffix="")
        T = SimpleGraphHelper.obtain_graph(args, suffix="T")
        return SubgraphFormula(G, [T], symmetric=False)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Graph isomorphimsm/automorphism formulas
"""

from cnfformula.cnf import CNF

from cnfformula.graphs import enumerate_vertices
from itertools import combinations, product


def _graph_isomorphism_var(u, v):
    """Standard variable name"""
    return "x_{{{0},{1}}}".format(u, v)


def GraphIsomorphism(G1, G2):
    """Graph Isomorphism formula

    The formula is the CNF encoding of the statement that two simple
    graphs G1 and G2 are isomorphic.

    Parameters
    ----------
    G1 : networkx.Graph
        an undirected graph object
    G2 : networkx.Graph
        an undirected graph object

    Returns
    -------
    A CNF formula which is satiafiable if and only if graphs G1 and G2
    are isomorphic.

    """
    description = "Graph isomorphism between (1) '{}' and (2) '{}'"
    description = description.format(G1.name, G2.name)

    F = CNF(description=description)

    U = enumerate_vertices(G1)
    V = enumerate_vertices(G2)
    var = _graph_isomorphism_var

    for (u, v) in product(U, V):
        F.add_variable(var(u, v))

    # Defined on both side
    for u in U:
        F.add_clause([(True, var(u, v)) for v in V], strict=True)

    for v in V:
        F.add_clause([(True, var(u, v)) for u in U], strict=True)

    # Injective on both sides
    for u in U:
        for v1, v2 in combinations(V, 2):
            F.add_clause([(False, var(u, v1)), (False, var(u, v2))],
                         strict=True)
    for v in V:
        for u1, u2 in combinations(U, 2):
            F.add_clause([(False, var(u1, v)), (False, var(u2, v))],
                         strict=True)

    # Edge consistency
    for u1, u2 in combinations(U, 2):
        for v1, v2 in combinations(V, 2):
            if G1.has_edge(u1, u2) != G2.has_edge(v1, v2):
                F.add_clause([(False, var(u1, v1)), (False, var(u2, v2))],
                             strict=True)
                F.add_clause([(False, var(u1, v2)), (False, var(u2, v1))],
                             strict=True)

    return F


def GraphAutomorphism(G):
    """Graph Automorphism formula

    The formula is the CNF encoding of the statement that a graph G
    has a nontrivial automorphism, i.e. an automorphism different from
    the idential one.

    Parameter
    ---------
    G : a simple graph

    Returns
    -------
    A CNF formula which is satiafiable if and only if graph G has a
    nontrivial automorphism.
    """
    description = "Graph automorphism formula for " + G.name
    F = GraphIsomorphism(G, G)
    F.header['description'] = description

    var = _graph_isomorphism_var

    F.add_clause([(False, var(u, u)) for u in enumerate_vertices(G)],
                 strict=True)

    return F



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of Thapen's size-width tradeoff formula
"""

from cnfgen.formula_helpers import FormulaHelper
from cnfgen.cmdline import positive_int

from cnfformula.families.cpls import CPLSFormula


class CPLSCmdHelper(FormulaHelper):
    """Command line helper for Thapen's size-width tradeoff formula"""

    name = 'cpls'
    description = 'Thapen\'s Coloured Polynomial Local Search formula'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for Thapen's size-width tradeoff formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('a',
                            metavar='<a>',
                            type=positive_int,
                            help="Number of levels")
        parser.add_argument(
            'b',
            metavar='<b>',
            type=positive_int,
            help=
            "Number of nodes per level in the graph (must be a power of two)")
        parser.add_argument('c',
                            metavar='<c>',
                            type=int,
                            help="Number of colours (must be a power of two)")

    @staticmethod
    def build_cnf(args):
        """Build Thapen's size-width tradeoff formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        return CPLSFormula(args.a, args.b, args.c)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of the pigeonhole principle formulas
"""

from cnfformula.cnf import CNF
from cnfformula.graphs import bipartite_sets

from cnfformula.graphs import neighbors
from itertools import combinations, product


def PigeonholePrinciple(pigeons, holes, functional=False, onto=False):
    """Pigeonhole Principle CNF formula

    The pigeonhole  principle claims  that no M  pigeons can sit  in N
    pigeonholes  without collision  if M>N.   The  counterpositive CNF
    formulation  requires  such mapping  to  be  satisfied. There  are
    different  variants of this  formula, depending  on the  values of
    `functional` and `onto` argument.

    - PHP: pigeon can sit in multiple holes
    - FPHP: each pigeon sits in exactly one hole
    - onto-PHP: pigeon can  sit in multiple holes, every  hole must be
                covered.
    - Matching: one-to-one bijection between pigeons and holes.

    Arguments:
    - `pigeon`: number of pigeons
    - `hole`:   number of holes
    - `functional`: add clauses to enforce at most one hole per pigeon
    - `onto`: add clauses to enforce that any hole must have a pigeon

    >>> print(PigeonholePrinciple(4,3).dimacs(export_header=False))
    p cnf 12 22
    1 2 3 0
    4 5 6 0
    7 8 9 0
    10 11 12 0
    -1 -4 0
    -1 -7 0
    -1 -10 0
    -4 -7 0
    -4 -10 0
    -7 -10 0
    -2 -5 0
    -2 -8 0
    -2 -11 0
    -5 -8 0
    -5 -11 0
    -8 -11 0
    -3 -6 0
    -3 -9 0
    -3 -12 0
    -6 -9 0
    -6 -12 0
    -9 -12 0
    """
    def var_name(p, h):
        return 'p_{{{0},{1}}}'.format(p, h)

    if functional:
        if onto:
            formula_name = "Matching"
        else:
            formula_name = "Functional pigeonhole principle"
    else:
        if onto:
            formula_name = "Onto pigeonhole principle"
        else:
            formula_name = "Pigeonhole principle"

    description = "{0} formula for {1} pigeons and {2} holes".format(
        formula_name, pigeons, holes)
    php = CNF(description=description)

    if pigeons < 0 or holes < 0:
        raise ValueError(
            "Number of pigeons and holes must both be non negative")

    mapping = php.unary_mapping(range(1, pigeons + 1),
                                range(1, holes + 1),
                                var_name=var_name,
                                injective=True,
                                functional=functional,
                                surjective=onto)

    for v in mapping.variables():
        php.add_variable(v)

    for c in mapping.clauses():
        php.add_clause_unsafe(c)

    return php


def GraphPigeonholePrinciple(graph, functional=False, onto=False):
    """Graph Pigeonhole Principle CNF formula

    The graph pigeonhole principle CNF formula, defined on a bipartite
    graph G=(L,R,E), claims that there is a subset E' of the edges such that
    every vertex on the left size L has at least one incident edge in E' and
    every edge on the right side R has at most one incident edge in E'.

    This is possible only if the graph has a matching of size |L|.

    There are different variants of this formula, depending on the
    values of `functional` and `onto` argument.

    - PHP(G):  each left vertex can be incident to multiple edges in E'
    - FPHP(G): each left vertex must be incident to exaclty one edge in E'
    - onto-PHP: all right vertices must be incident to some vertex
    - matching: E' must be a perfect matching between L and R

    Arguments:
    - `graph` : bipartite graph
    - `functional`: add clauses to enforce at most one edge per left vertex
    - `onto`: add clauses to enforce that any right vertex has one incident edge


    Remark: the graph vertices must have the 'bipartite' attribute
    set. Left vertices must have it set to 0 and the right ones to
    1. A KeyException is raised otherwise.

    """
    def var_name(p, h):
        return 'p_{{{0},{1}}}'.format(p, h)

    if functional:
        if onto:
            formula_name = "Graph matching"
        else:
            formula_name = "Graph functional pigeonhole principle"
    else:
        if onto:
            formula_name = "Graph onto pigeonhole principle"
        else:
            formula_name = "Graph pigeonhole principle"

    description = "{0} formula on {1}".format(formula_name, graph.name)
    gphp = CNF(description=description)

    Left, Right = bipartite_sets(graph)

    mapping = gphp.unary_mapping(Left,
                                 Right,
                                 sparsity_pattern=graph,
                                 var_name=var_name,
                                 injective=True,
                                 functional=functional,
                                 surjective=onto)

    for v in mapping.variables():
        gphp.add_variable(v)

    for c in mapping.clauses():
        gphp.add_clause_unsafe(c)

    return gphp


def BinaryPigeonholePrinciple(pigeons, holes):
    """Binary Pigeonhole Principle CNF formula

    The pigeonhole principle claims that no M pigeons can sit in
    N pigeonholes without collision if M>N. This formula encodes the
    principle using binary strings to identify the holes.

    Parameters
    ----------
    pigeon : int
       number of pigeons
    holes : int
       number of holes
    """

    description = "Binary Pigeonhole Principle for {0} pigeons and {1} holes".format(
        pigeons, holes)
    bphp = CNF(description=description)

    if pigeons < 0 or holes < 0:
        raise ValueError(
            "Number of pigeons and holes must both be non negative")

    bphpgen = bphp.binary_mapping(range(1, pigeons + 1),
                                  range(1, holes + 1),
                                  injective=True)

    for v in bphpgen.variables():
        bphp.add_variable(v)

    for c in bphpgen.clauses():
        bphp.add_clause_unsafe(c)

    return bphp
